//-------------------------------------------------------------------------------------
//
// Copyright 2007 Intel Corporation
// All Rights Reserved
//
// Permission is granted to use, copy, distribute and prepare derivative works of this
// software for any purpose and without fee, provided, that the above copyright notice
// and this statement appear in all copies.  Intel makes no representations about the
// suitability of this software for any purpose.  THIS SOFTWARE IS PROVIDED "AS IS."
// INTEL SPECIFICALLY DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, AND ALL LIABILITY,
// INCLUDING CONSEQUENTIAL AND OTHER INDIRECT DAMAGES, FOR THE USE OF THIS SOFTWARE,
// INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PROPRIETARY RIGHTS, AND INCLUDING THE
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  Intel does not
// assume any responsibility for any errors which may appear in this software nor any
// responsibility to update it.
//
// Portions  fBm, HybridMultiFractal, and RidgedMultiFractal code used with permission
//	"Texturing and Modeling: A Procedural Approach", Ebert, Musgrave, Peachey, Perlin,
//	Worley, 1993, Academic Press, Inc.
//	Copyright 1994 F. Kenton Musgrave
//
//-------------------------------------------------------------------------------------

#include "DXUT.h"
#include "FractalMesh.h"
#include "Fractal.h"
#include "Perlin.h"


#define X_TRANSLATE_OFFSET 5.0f	// x-axis offset to adjust the mesh for viewing

//--------------------------------------------------------------------------------------
// Create and initialize the mesh
//--------------------------------------------------------------------------------------
FractalMesh::FractalMesh( IDirect3DDevice9* pDevice, 
						 int iNumRows, 
						 int iNumCols, 
						 FRACTAL_ALGORITHM defaultAlgorithm ) : m_fractalAlgorithm( defaultAlgorithm )
{
	Initialize( pDevice, iNumRows, iNumCols );
}


//--------------------------------------------------------------------------------------
// Frees DirectX resources
//--------------------------------------------------------------------------------------
FractalMesh::~FractalMesh()
{
	DeleteMesh();
}


//--------------------------------------------------------------------------------------
// Initialize the mesh, face normals, and other basic setup
//--------------------------------------------------------------------------------------
void FractalMesh::Initialize( IDirect3DDevice9 *pDevice, int iNumRows, int iNumCols )
{
	m_bUseFastNormalLookup = true;

	m_iNumRows = iNumRows;
	m_iNumCols = iNumCols;
	m_pIB = NULL;
	m_pVB = NULL;

	m_pNormalLookupTable = NULL;
	m_pNeighborList = NULL;
	m_pFaceNormals = NULL;

	m_pIBD3D = NULL;
	m_pVBD3D = NULL;
	m_pDevice = pDevice;

	m_bFirstDerivativeThru = false;
	m_bFirstNormalLookupThru = false;

	m_FractalExponentArray = NULL;
	m_firstFractalCalc = false;

	m_fMaxElevation = 0;

	GridGen();
}


//--------------------------------------------------------------------------------------
// Compute a bounding sphere for use in setting camera positions
//--------------------------------------------------------------------------------------
void FractalMesh::ComputeBoundingSphere( CVector3 *ctr, float *radius )
{
	// Walk the vertex buffer, determine the farthest point from the origin
	//	distance from origin = v2-v1
	//	determine the origin, the average of all of the points
	CVector3 center;
	ComputeOrigin( &center );

	// Calculate the distance from each point to the center of the mesh
	//	keep track of the furthest distance
	int curFurthestPtIndex = -1;
	float curFurthestDistance = -1.0f;
	float tempDist = -1.0f;
	CVector3 temp;
	temp.Init( 0.0f, 0.0f, 0.0f );

	for( int i=0; i<m_iNumRows*m_iNumCols; i++ )
	{
		temp.x = m_pVB[i].x - center.x;
		temp.y = m_pVB[i].y - center.y;
		temp.z = m_pVB[i].z - center.z;
		tempDist = sqrt( temp.DotSelf() );
		if( tempDist > curFurthestDistance )
		{
			//new furthest distance
			curFurthestDistance = tempDist;
			curFurthestPtIndex = i;
		}
	}
	*radius = curFurthestDistance;
	ctr->Init( temp.x, temp.y, temp.z );
}


//--------------------------------------------------------------------------------------
// Locate the origin for the bounding box computation
//--------------------------------------------------------------------------------------
void FractalMesh::ComputeOrigin( CVector3 *ctr )
{
	CVector3 sum;
	sum.Init( 0.0f, 0.0f, 0.0f );
	// Walk all of the vertices
	int i = 0;
	for( i=0; i<(m_iNumRows)*(m_iNumCols); i++ )
	{
		sum.x += m_pVB[i].x;
		sum.y += m_pVB[i].y;
		sum.z += m_pVB[i].z;
	}

	_ASSERT( i > 0 );
	sum.DivConst( (float)i );

	ctr->Init( sum.x, sum.y, sum.z );
}


//--------------------------------------------------------------------------------------
// Get accessor
//--------------------------------------------------------------------------------------
LPDIRECT3DVERTEXBUFFER9 FractalMesh::GetD3DVB()
{
	return m_pVBD3D;
}


//--------------------------------------------------------------------------------------
// Get accessor
//--------------------------------------------------------------------------------------
LPDIRECT3DINDEXBUFFER9 FractalMesh::GetD3DIB()
{
	return m_pIBD3D;
}


//--------------------------------------------------------------------------------------
// Vertex neighbor list used in vertex normal calculations
//--------------------------------------------------------------------------------------
void FractalMesh::InitNeighborList()
{
	if( m_pNeighborList == NULL )
	{
		m_pNeighborList = (int *)malloc( sizeof(int) * m_iNumRows * m_iNumCols * 6 );
	}

	for( int i=0; i<m_iNumRows*m_iNumCols; i++ )
	{
		for( int j=0; j<6; j++ )
		{
			m_pNeighborList[i*6+j] = -1;
		}
	}
}


//--------------------------------------------------------------------------------------
// For vertex normal calculations
//--------------------------------------------------------------------------------------
void FractalMesh::AddToNeighborList( int me, int neighbor )
{
	for( int i=0; i<6; i++ )
	{
		if( m_pNeighborList[me*6+i] == neighbor )
		{ // Already in list
			return;
		}
		if( m_pNeighborList[me*6+i] == -1 )
		{
			m_pNeighborList[me*6+i] = neighbor;
			return;
		}
	}
}


//--------------------------------------------------------------------------------------
// Check if the vertex is a neighbor of another vertex for normal calculations
//--------------------------------------------------------------------------------------
int FractalMesh::IsNeighborOf( int a, int b )
{
	for( int i=0; i<6; i++ )
	{
		if( m_pNeighborList[a*6+i] == -1 )
		{
			return FALSE;
		}
		else if( m_pNeighborList[a*6+i] == b )
		{
			return TRUE;
		}
	}
	// No neighbor found
	return FALSE;
}


//--------------------------------------------------------------------------------------
// Allocate DirectX resources for the mesh and calculate normals
//--------------------------------------------------------------------------------------
void FractalMesh::GridGen()
{
	InitNeighborList();
	InitFaceNormals();

	GenVB();
	GenIB();
	GenNormalLookupTable();
	CreateFaceNormals();

	if( m_bUseFastNormalLookup )
	{
		CalcPerVertexNormalsFastLookup();
	}
	else
	{
		CalcPerVertexNormals();
	}

	LoadD3D_VB();
	LoadD3D_IB();
}


//--------------------------------------------------------------------------------------
// Recalculates the terrain with the selected fractal algorithm
//--------------------------------------------------------------------------------------
void FractalMesh::ReComputeTerrain( FRACTAL_ALGORITHM fractalAlgorithm )
{
	m_fractalAlgorithm = fractalAlgorithm;
	ComputeTerrain();

	CreateFaceNormals();

	//DAR:  Don't have to worry about the derivative method, because the technique which sums the wave derivatives
	//      doesn't use the face normals at all.
	if( m_bUseFastNormalLookup )
	{
		CalcPerVertexNormalsFastLookup();
	}
	else
	{
		CalcPerVertexNormals();
	}

	ReloadVB();
	ReloadIB();
}

//--------------------------------------------------------------------------------------
// Calculates the terrain with the selected fractal algorithm, normals are calculated 
//	in the callee after surface perturbation
//--------------------------------------------------------------------------------------
void FractalMesh::ComputeTerrain()
{
	float fCurColVal = 0.0f;
	float fCurRowVal = 0.0f;
	m_fMaxElevation = 0.0f;

	float fStepSizeAcrossRow = 20.0f / float( m_iNumCols-1 );
	float fStepSizeDownCol = 20.0f / float( m_iNumRows-1 );

	float vertex[3];
	if( m_pVB != NULL )
	{
		m_firstFractalCalc = true;
		for( int iCurRow = 0;iCurRow < m_iNumRows; iCurRow++ )
		{
			for( int iCurCol = 0;iCurCol < m_iNumCols; iCurCol++ )
			{
				vertex[0] = (fCurColVal);
				vertex[1] = fCurRowVal;
				vertex[2] = 0.0f;

				switch( m_fractalAlgorithm )
				{
				case FBM:
					{
						m_pVB[iCurRow*m_iNumCols+iCurCol].z = (float)(fBm( vertex, m_H, m_Lacunarity, m_Octaves ) );

						if( m_pVB[iCurRow*m_iNumCols+iCurCol].z > m_fMaxElevation )
						{
							m_fMaxElevation = m_pVB[iCurRow*m_iNumCols+iCurCol].z;
						}

						break;
					}
				case HYBRIDMULTIFRACTAL:
					{
						m_pVB[iCurRow*m_iNumCols+iCurCol].z = (float)(HybridMultiFractal( vertex, m_H, m_Lacunarity, m_Octaves, m_Offset ) );

						if( m_pVB[iCurRow*m_iNumCols+iCurCol].z > m_fMaxElevation )
						{
							m_fMaxElevation = m_pVB[iCurRow*m_iNumCols+iCurCol].z;
						}

						break;
					}
				case RIDGEDMULTIFRACTAL:
					{
						m_pVB[iCurRow*m_iNumCols+iCurCol].z = (float)(RidgedMultiFractal( vertex, m_H, m_Lacunarity, m_Octaves, m_Offset, m_Gain ) );

						if( m_pVB[iCurRow*m_iNumCols+iCurCol].z > m_fMaxElevation )
						{
							m_fMaxElevation = m_pVB[iCurRow*m_iNumCols+iCurCol].z;
						}

						break;
					}
				default:
					{
						assert( false );
					}
				}

				fCurColVal += fStepSizeAcrossRow;
			}
			fCurColVal = 0.0f;
			fCurRowVal += fStepSizeDownCol;
		}

		switch( m_fractalAlgorithm )
		{
		case FBM:
		case HYBRIDMULTIFRACTAL:
		case RIDGEDMULTIFRACTAL:
			{
				if( m_FractalExponentArray )
				{
					free( m_FractalExponentArray );
					m_FractalExponentArray = NULL;
				}
				m_firstFractalCalc = true;
				break;
			}
		default:
			{
				assert( false );
				break;
			}
		}
	}
}


//--------------------------------------------------------------------------------------
// Initialize face normals for the mesh
//--------------------------------------------------------------------------------------
void FractalMesh::InitFaceNormals()
{
	m_pFaceNormals = new CVector3[(m_iNumRows-1)*(m_iNumCols-1)*2];

	for( int i=0; i<(m_iNumRows-1)*(m_iNumCols-1)*2; i++ )
	{
		m_pFaceNormals[i].Init( 0.0f, 0.0f, 0.0f );
	}
}


//--------------------------------------------------------------------------------------
// Create face normals for the mesh
//--------------------------------------------------------------------------------------
void FractalMesh::CreateFaceNormals()
{
	// Loop over how many faces there are
	int j=0;
	for( int i=0; i<(m_iNumRows-1)*(m_iNumCols-1)*6; i+=3 )
	{
		// Create vector v-a, v-b
		CVector3 v2minusv0;
		CVector3 v1minusv0;
		v1minusv0.Init( m_pVB[m_pIB[i+1]].x - m_pVB[m_pIB[i]].x, m_pVB[m_pIB[i+1]].y - m_pVB[m_pIB[i]].y, m_pVB[m_pIB[i+1]].z - m_pVB[m_pIB[i]].z );
		v2minusv0.Init( m_pVB[m_pIB[i+2]].x - m_pVB[m_pIB[i]].x, m_pVB[m_pIB[i+2]].y - m_pVB[m_pIB[i]].y, m_pVB[m_pIB[i+2]].z - m_pVB[m_pIB[i]].z );
		v1minusv0.Cross( &v2minusv0, &m_pFaceNormals[j] );
		m_pFaceNormals[j].Normalize();
		j++;
	}
}


//--------------------------------------------------------------------------------------
// Called from GridGen to allocate a vertex buffer
//--------------------------------------------------------------------------------------
void FractalMesh::GenVB()
{
	float fStepSizeAcrossRow = 20.0f / float( m_iNumCols-1 );
	float fStepSizeDownCol = 20.0f / float( m_iNumRows-1 );

	float fCurColVal = 0.0f;
	float fCurRowVal = 0.0f;

	if( m_pVB != NULL )
	{
		free( m_pVB );
		m_pVB = NULL;
	}
	m_pVB = ( CUSTOM_VERT_POS_NORMAL_TEXTURE0 *)malloc(m_iNumCols*m_iNumRows*sizeof(CUSTOM_VERT_POS_NORMAL_TEXTURE0) );

	for( int iCurRow = 0;iCurRow < m_iNumRows; iCurRow++ )
	{
		for( int iCurCol = 0;iCurCol < m_iNumCols; iCurCol++ )
		{
			// Flip these if you want row major
			m_pVB[iCurRow*m_iNumCols+iCurCol].x = (fCurColVal-5) - X_TRANSLATE_OFFSET;
			m_pVB[iCurRow*m_iNumCols+iCurCol].y = fCurRowVal-3;
			m_pVB[iCurRow*m_iNumCols+iCurCol].z = 0.0f;
			// Texture coords
			m_pVB[iCurRow*m_iNumCols+iCurCol].tu = fCurColVal;
			m_pVB[iCurRow*m_iNumCols+iCurCol].tv = fCurRowVal;
			fCurColVal += fStepSizeAcrossRow;
		}
		fCurColVal = 0.0f;
		fCurRowVal += fStepSizeDownCol;
	}
}


//--------------------------------------------------------------------------------------
// Allocate DirectX resources for the vertex buffer
//--------------------------------------------------------------------------------------
void FractalMesh::LoadD3D_VB()
{
	int iResult = -1;

	iResult = m_pDevice->CreateVertexBuffer( m_iNumRows*m_iNumCols*sizeof(CUSTOM_VERT_POS_NORMAL_TEXTURE0),
		D3DUSAGE_WRITEONLY,
		D3D_FVF_CUSTOMVERTEX_POS_NORMAL_TEXTURE0,
		D3DPOOL_DEFAULT,
		&m_pVBD3D,
		NULL );

	assert( iResult == D3D_OK ); // Create VB Failed

	// Fill VB
	VOID *pVerts = NULL;
	iResult = m_pVBD3D->Lock( 0, m_iNumRows*m_iNumCols*sizeof(CUSTOM_VERT_POS_NORMAL_TEXTURE0), (void **)&pVerts, 0 );

	assert( iResult == D3D_OK ); // Fill VB fail

	memcpy( pVerts, m_pVB, m_iNumRows*m_iNumCols*sizeof(CUSTOM_VERT_POS_NORMAL_TEXTURE0) );
	m_pVBD3D->Unlock();
}


//--------------------------------------------------------------------------------------
// Reallocate DirectX resources for the vertex buffer
//--------------------------------------------------------------------------------------
void FractalMesh::ReloadVB()
{
	int iResult;
	// Fill VB
	VOID *pVerts = NULL;
	iResult = m_pVBD3D->Lock( 0, m_iNumRows*m_iNumCols*sizeof(CUSTOM_VERT_POS_NORMAL_TEXTURE0), (void **)&pVerts, 0 );
	
	assert( iResult == D3D_OK ); // Lock VB failed
	
	memcpy( pVerts, m_pVB, m_iNumRows*m_iNumCols*sizeof(CUSTOM_VERT_POS_NORMAL_TEXTURE0) );
	m_pVBD3D->Unlock();
}


//--------------------------------------------------------------------------------------
// Reallocate DirectX resources for the index buffer
//--------------------------------------------------------------------------------------
void FractalMesh::ReloadIB()
{
	int iResult = -1;
	int iBufSize = sizeof( short int ) * ( (m_iNumRows-1) * (m_iNumCols-1) * 6 );

	// Fill IB
	VOID *pIndices = NULL;
	iResult = m_pIBD3D->Lock( 0, iBufSize, (void **)&pIndices, 0 );
	
	assert( iResult == D3D_OK ); // Lock IB failed

	memcpy( pIndices, m_pIB, iBufSize );
	m_pIBD3D->Unlock();
}


//--------------------------------------------------------------------------------------
// Allocate DirectX resources for the index buffer
//--------------------------------------------------------------------------------------
void FractalMesh::LoadD3D_IB()
{
	int iResult = -1;
	int iBufSize = sizeof( short int ) * ( (m_iNumRows-1) * (m_iNumCols-1) * 6 );

	iResult = m_pDevice->CreateIndexBuffer( iBufSize, D3DUSAGE_WRITEONLY, D3DFMT_INDEX16, D3DPOOL_DEFAULT, &m_pIBD3D, NULL );

	assert( iResult == D3D_OK ); // Create IB failed
	
	// Fill IB
	VOID *pIndices = NULL;
	iResult = m_pIBD3D->Lock( 0, iBufSize, (void **)&pIndices, 0 );
	
	assert( iResult == D3D_OK ); // Lock IB failed
	
	memcpy( pIndices, m_pIB, iBufSize );
	m_pIBD3D->Unlock();
}

//--------------------------------------------------------------------------------------
// Create a fast normal lookup table
//--------------------------------------------------------------------------------------
void FractalMesh::GenNormalLookupTable()
{
	// We're going to create a table that lists, for each vertex, the
	// triangles which that vertex is a part of.

	if( m_pNormalLookupTable != NULL )
	{
		free( m_pNormalLookupTable );
		m_pNormalLookupTable = NULL;
	}

	// Each vertex may be a part of up to 6 triangles in the grid, so
	// create a buffer to hold a pointer to the normal of each neighbor.
	int buffersize = m_iNumRows*m_iNumCols*6;

	m_pNormalLookupTable = (int *)malloc( sizeof(void *) * buffersize );
	for( int i=0; i<buffersize; i++ )
		m_pNormalLookupTable[i] = -1;

	// Now that the table is initialized, populate it with the triangle data.

	// For each triangle
	//   For each vertex in that triangle
	//     Append the triangle number to lookuptable[vertex]
	for( int i=0; i < 2*(m_iNumRows-1)*(m_iNumCols-1); i++ )
	{
		for( int j=0; j<3; j++ )
		{
			// Find the next empty slot in the vertex's lookup table "slot"
			for( int k=0; k<6; k++ )
			{
				int vertex = m_pIB[i*3+j];
				if( m_pNormalLookupTable[vertex*6+k] == -1 )
				{
					m_pNormalLookupTable[vertex*6+k] = i;
					break;
				}
			}
		}  // For each vertex that is part of the current triangle
	} // For each triangle
}


//--------------------------------------------------------------------------------------
// Generate the index buffer for the mesh
//--------------------------------------------------------------------------------------
void FractalMesh::GenIB()
{
	int i=0;

	if( m_pIB != NULL )
	{
		free( m_pIB );
		m_pIB = NULL;
	}

	m_pIB = (short int *)malloc( sizeof(short int) * (m_iNumRows-1) * (m_iNumCols-1) * 6 );

	i = 0;
	int iCurIndex = 0;
	for( int iCurRow = 0; iCurRow < m_iNumRows-1; iCurRow++ )
	{
		for( int iCurCol = 0; iCurCol < m_iNumCols-1; iCurCol++ )
		{
			int tri = ( iCurRow*m_iNumCols ) + iCurCol;
			iCurIndex = i;
			m_pIB[iCurIndex++] = (short)tri;
			m_pIB[iCurIndex++] = (short)(tri+m_iNumCols+1);
			m_pIB[iCurIndex++] = (short)(tri+m_iNumCols);
			m_pIB[iCurIndex++] = (short)tri;
			m_pIB[iCurIndex++] = (short)(tri+1);
			m_pIB[iCurIndex++] = (short)(tri+m_iNumCols+1);

			// Add the indices to the neighborlist
			// This was calculated by looking at a grid and the neighbors
			// of a particular triangle

			AddToNeighborList( m_pIB[i+0], m_pIB[i+1] );
			AddToNeighborList( m_pIB[i+0], m_pIB[i+2] );
			AddToNeighborList( m_pIB[i+0], m_pIB[i+4] );

			AddToNeighborList( m_pIB[i+4], m_pIB[i+0] );
			AddToNeighborList( m_pIB[i+4], m_pIB[i+1] );

			AddToNeighborList( m_pIB[i+1], m_pIB[i+2] );
			AddToNeighborList( m_pIB[i+1], m_pIB[i+4] );
			AddToNeighborList( m_pIB[i+1], m_pIB[i] );

			AddToNeighborList( m_pIB[i+2], m_pIB[i+0] );
			AddToNeighborList( m_pIB[i+2], m_pIB[i+1] );

			i+=6;
		}
	}
}


//--------------------------------------------------------------------------------------
// Compute vertex normals from the fast normal lookup table
//--------------------------------------------------------------------------------------
void FractalMesh::CalcPerVertexNormalsFastLookup()
{
	// First, calculate the face normals for each triangle.
	CreateFaceNormals();

	// For each vertex, sum the normals indexed by the normal lookup table.
	int j = 0;
	for( int i=0; i < (m_iNumCols * m_iNumRows); i++ )
	{
		CVector3 avgNormal;
		avgNormal.Init( 0.0f, 0.0f, 0.0f );

		// Find all the triangles that this vertex is a part of.
		for( j=0; j<6; j++ )
		{
			int triIndex;
			triIndex = m_pNormalLookupTable[i*6+j];

			// If the triangle index is valid, get the normal and average it in.
			if( triIndex != -1 )
			{
				CVector3 normal;
				LookupNormal( triIndex, &normal );
				avgNormal.AddVec3( &normal );
			}
			else
				break;
		}

		// Complete the averaging step by dividing & normalizing.
		_ASSERT( j > 0 );
		avgNormal.DivConst( (float)(j) );
		avgNormal.Normalize();

		// Set the vertex normal to this new normal.
		m_pVB[i].nx = avgNormal.x;
		m_pVB[i].ny = avgNormal.y;
		m_pVB[i].nz = avgNormal.z;

	}  // For each vertex
}


//--------------------------------------------------------------------------------------
// Calcute vertex normals
//--------------------------------------------------------------------------------------
void FractalMesh::CalcPerVertexNormals()
{
	//doubly nested loop: start at top of NeighborList (nx6 array)
	//if v != -1, get value, called a
	//for each value after that, get the next one, call it b (we'll go through all of them due to for loop in later iterations)
	//go to a neighbor list, if b in a's neighbor list, calculate a normal doing:
	//store in temp vector = b-v cross a-v
	//go to next vertex (will be b)

	CreateFaceNormals();

	int v = -1;
	for( int i=0; i<m_iNumCols*m_iNumRows; i++ )
	{
		CVector3 avgNormal;
		avgNormal.Init( 0.0f, 0.0f, 0.0f );

		v = i;
		float divisor = 1.0f;
		for( int j=0; j<6; j++ )
		{
			if( m_pNeighborList[i*6+j] != -1 )
			{
				int a = m_pNeighborList[i*6+j];
				// We only need to go from j, everythign before been checked!
				for( int k=j+1; k<6; k++ )
				{
					if( m_pNeighborList[i*6+k] != -1 )
					{
						int b = m_pNeighborList[i*6+k];
						// Check to see if its a neighbor of a
						int result = IsNeighborOf( a, b );
						// If they are neighbors compute the normal
						if( result == TRUE )
						{
							// Lookup triangle index
							int triIndex = LookupTriIndex( v, a, b );

							// Get its normal
							CVector3 normal;
							LookupNormal( triIndex,&normal );
							// Already normalized, just compute avg.
							avgNormal.AddVec3( &normal );
							divisor += 1.0f;
						}
					}
					else
					{ // End of list
						break;
					}
				} // for loop over the rest of the neighbor list (k=j)
			}
		}  // for loop over neighbors (j)


		// Average the normals
		avgNormal.DivConst( divisor );
		avgNormal.Normalize();
		// Set the vertex normal to this new normal
		m_pVB[i].nx = avgNormal.x;
		m_pVB[i].ny = avgNormal.y;
		m_pVB[i].nz = avgNormal.z;

	} //for loop over the verts (i)
}


//--------------------------------------------------------------------------------------
// Retrieve the polygon index from the IB given the triangle vertices
//--------------------------------------------------------------------------------------
int FractalMesh::LookupTriIndex( int v1, int v2, int v3 )
{
	int index = 0;
	for( int i=0; i<(m_iNumRows-1)*(m_iNumCols-1)*6; i+=3 )
	{
		if( m_pIB[i] == v1 || m_pIB[i+1] == v1 || m_pIB[i+2] == v1 )
			if( m_pIB[i] == v2 || m_pIB[i+1] == v2 || m_pIB[i+2] == v2 )
				if( m_pIB[i] == v3 || m_pIB[i+1] == v3 || m_pIB[i+2] == v3 )
					return index;
		index++;
	}
	return -1;
}


//--------------------------------------------------------------------------------------
// Retrieve the face normal for the specified vertex
//--------------------------------------------------------------------------------------
void FractalMesh::LookupNormal( int v1, CVector3 *res )
{
	res->x = m_pFaceNormals[v1].x;
	res->y = m_pFaceNormals[v1].y;
	res->z = m_pFaceNormals[v1].z;
}


//--------------------------------------------------------------------------------------
// Free allocated DirectX resources for the mesh
//--------------------------------------------------------------------------------------
void FractalMesh::DeleteMesh()
{
	if( m_pNormalLookupTable )
	{
		free( m_pNormalLookupTable );
		m_pNormalLookupTable = NULL;
	}
	if( m_pIB )
	{
		free( m_pIB );
		m_pIB = NULL;
	}
	if( m_pVB )
	{
		free( m_pVB );
		m_pVB = NULL;
	}

	if( m_pIBD3D != NULL )
	{
		m_pIBD3D->Release();
		m_pIBD3D = NULL;
	}
	if( m_pVBD3D != NULL )
	{
		m_pVBD3D->Release();
		m_pVBD3D = NULL;
	}
	if( m_pVBD3D != NULL )
	{
		m_pVBD3D->Release();
		m_pVBD3D = NULL;
	}
	if( m_pNeighborList != NULL )
	{
		free( m_pNeighborList );
		m_pNeighborList = NULL;
	}
	if( m_pFaceNormals != NULL )
	{
		delete m_pFaceNormals;
		m_pFaceNormals = NULL;
	}

	m_iNumRows = -1;
	m_iNumCols = -1;
}


//--------------------------------------------------------------------------------------
// DXUT registered GUI message
//--------------------------------------------------------------------------------------
void FractalMesh::OnLostDevice()
{
	DeleteMesh();
}


//--------------------------------------------------------------------------------------
// DXUT registered GUI message
//--------------------------------------------------------------------------------------
void FractalMesh::OnResetDevice(IDirect3DDevice9 *pDevice, int iNumRows, int iNumCols )
{
	Initialize( pDevice,iNumRows, iNumCols );
}


//--------------------------------------------------------------------------------------
// Necessary intermediate rendering step
//--------------------------------------------------------------------------------------
void FractalMesh::CopyVBToRenderVB( void )
{
	ReloadVB();
}


//--------------------------------------------------------------------------------------
// Get the relative highest point int the mesh to serve in the pixel shader's
//	region calculation.  This removes the dependency on fixed values which can
//	skew results of some fractal and multifractal settings that provide terrain with
//	properties that more closely resemble terrain as viewed from higher or lower viewpoint
//	perspectives.
//--------------------------------------------------------------------------------------
float FractalMesh::GetMaxElevation( void )
{
	return m_fMaxElevation;
}


//--------------------------------------------------------------------------------------
// Simple fractional Brownian motion adapted from "Texturing and Modeling: A Procedural 
//	Approach"
//--------------------------------------------------------------------------------------
double FractalMesh::fBm( float vertex[], double H, double lacunarity, double octaves )
{
	double value, frequency, remainder;
	int i;

	if( m_firstFractalCalc )
	{
		m_FractalExponentArray = (double *)malloc( (size_t)(octaves+1) * sizeof(double) );
		frequency = 1.0;
		for (i=0; i<=(int)octaves; i++)
		{
			m_FractalExponentArray[i] = pow( frequency, -H );
			frequency *= lacunarity;
		}

		m_firstFractalCalc = false;
	}

	value = 0.0;
	frequency = 1.0;

	for (i=0; i<(int)octaves; i++)
	{
		value += noise3( vertex ) * m_FractalExponentArray[i];
		vertex[0] *= (float)lacunarity;
		vertex[1] *= (float)lacunarity;
		vertex[2] *= (float)lacunarity;
	}

	remainder = octaves - (int)octaves;
	if (remainder)
		value += remainder * noise3( vertex ) * m_FractalExponentArray[i];

	if( value < 0.0 )
		value = 0.0f;

	return value;
}


//--------------------------------------------------------------------------------------
// Hybrid fractional Brownian motion multifractal adapted from "Texturing and Modeling: 
//	A Procedural Approach"
//--------------------------------------------------------------------------------------
double FractalMesh::HybridMultiFractal( float vertex[], double H, double lacunarity, 
									   double octaves, double offset )
{
	double frequency, result, signal, weight, remainder;
	int i;

	if( octaves < 1.0f )
		octaves = 1.0f;

	if( m_firstFractalCalc )
	{
		m_FractalExponentArray = (double *)malloc( (size_t)(octaves + 1) * sizeof(double) );
		frequency = 1.0;
		for (i = 0; i <=(int)octaves; i++ )
		{
			m_FractalExponentArray[i] = pow( frequency, -H );
			frequency *= lacunarity;
		}

		m_firstFractalCalc = false;
	}

	result = ( noise3( vertex ) + offset ) * m_FractalExponentArray[0];
	weight = result;

	vertex[0] *= (float)lacunarity;
	vertex[1] *= (float)lacunarity;
	vertex[2] *= (float)lacunarity;

	for (i=1; i<(int)octaves; i++)
	{
		if( weight > 1.0 )
			weight = 1.0;

		signal = ( noise3( vertex ) + offset ) * m_FractalExponentArray[i];
		result += weight * signal;

		weight *= signal;

		vertex[0] *= (float)lacunarity;
		vertex[1] *= (float)lacunarity;
		vertex[2] *= (float)lacunarity;
	}

	remainder = octaves - (int)octaves;
	if (remainder)
		result += remainder * noise3( vertex ) * m_FractalExponentArray[i];

	if( result < 0.0 )
		result = 0.0f;

	return result;
}


//--------------------------------------------------------------------------------------
// Ridged fractional Brownian motion multifractal adapted from "Texturing and Modeling: 
//	A Procedural Approach"
//	Intends to generate perturbations relative to sea level in hopes of creating more 
//	deeply defined ridglines
//--------------------------------------------------------------------------------------
double FractalMesh::RidgedMultiFractal( float vertex[], double H, double lacunarity, 
									   double octaves, double offset, double gain )
{
	double result, frequency, signal, weight;
	int i;

	if( octaves < 1.0f )
		octaves = 1.0f;

	if( m_firstFractalCalc )
	{
		m_FractalExponentArray = (double *)malloc( (size_t)(octaves+1)*sizeof(double) );
		frequency = 1.0;
		for (i=0; i<=(int)octaves; i++)
		{
			m_FractalExponentArray[i] = pow( (double)frequency, double(-H) );
			frequency *= lacunarity;
		}
		m_firstFractalCalc = false;
	}

	signal = noise3( vertex );
	if (signal < 0.0)
		signal = -signal;
	signal = offset - signal;
	signal *= signal;
	result = signal;
	weight = 1.0;
	for (i =1; i <(int)octaves; i++)
	{
		vertex[0] *= (float)lacunarity;
		vertex[1] *= (float)lacunarity;
		vertex[2] *= (float)lacunarity;

		weight = signal * gain;
		if (weight > 1.0)
			weight = 1.0;
		if (weight < 0.0)
			weight = 0.0;
		signal = noise3( vertex );
		if (signal < 0.0)
			signal = -signal;
		signal = offset - signal;
		signal *= signal;
		signal *= weight;
		result += signal * m_FractalExponentArray[i];
	}

	if( result < 0.0 )
		result = 0.0f;

	return result;
}