//-------------------------------------------------------------------------------------
//
// Copyright 2007 Intel Corporation
// All Rights Reserved
//
// Permission is granted to use, copy, distribute and prepare derivative works of this
// software for any purpose and without fee, provided, that the above copyright notice
// and this statement appear in all copies.  Intel makes no representations about the
// suitability of this software for any purpose.  THIS SOFTWARE IS PROVIDED "AS IS."
// INTEL SPECIFICALLY DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, AND ALL LIABILITY,
// INCLUDING CONSEQUENTIAL AND OTHER INDIRECT DAMAGES, FOR THE USE OF THIS SOFTWARE,
// INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PROPRIETARY RIGHTS, AND INCLUDING THE
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  Intel does not
// assume any responsibility for any errors which may appear in this software nor any
// responsibility to update it.
//
//-------------------------------------------------------------------------------------

#ifndef FractalMesh_H
#define FractalMesh_H

//class to generate a CGrid

#include "CUtil.h"
#include <stdlib.h>
#include <crtdbg.h>

#include <d3d9.h>

//NOTE:  Must be compiled with multithreaded MSVC++ libraries!
#include <process.h>

#include "DXUT.h"
#include "D3DDataTypes.h"


//--------------------------------------------------------------------------------------
// Supported fractal algorithms
//--------------------------------------------------------------------------------------
enum FRACTAL_ALGORITHM { FBM, HYBRIDMULTIFRACTAL, RIDGEDMULTIFRACTAL };

#define TRUE 1
#define FALSE 0

//--------------------------------------------------------------------------------------
// Helper macros to build member functions that access member variables
//--------------------------------------------------------------------------------------
#define FM_SET_ACCESSOR( x, y )       inline void Set##y( x t )   { m_##y = t; };
#define FM_GET_ACCESSOR( x, y )       inline x Get##y()           { return m_##y; };
#define FM_GET_SET_ACCESSOR( x, y )   FM_SET_ACCESSOR( x, y ) FM_GET_ACCESSOR( x, y )

class FractalMesh {
public:
	FractalMesh(IDirect3DDevice9 *pDevice, int iNumRows, int iNumCols, FRACTAL_ALGORITHM defaultAlgorithm);
	~FractalMesh();

	void Initialize(IDirect3DDevice9 *pDevice, int iNumRows, int iNumCols);
	void InitializeWave(int index, float wavelength, float amplitude, float speed, CVector3 *direction, float kexp);
	void GridGen();
	void ComputeTerrain();
	void ReComputeTerrain( FRACTAL_ALGORITHM fractalAlgorithm );

	int LookupTriIndex(int v1, int v2, int v3);
	void InitFaceNormals();
	void CreateFaceNormals();
	void CalcPerVertexNormals();
	void CalcPerVertexNormalsFastLookup();
	void CalcPerVertexNormalsWaveDerivative( float time );

	void LookupNormal(int v1, CVector3 *res);
	void ComputeBoundingSphere(CVector3 *ctr, float *radius);
	void ComputeOrigin(CVector3 *ctr);

	void OnLostDevice();
	void OnResetDevice(IDirect3DDevice9 *pDevice, int iNumRows, int iNumCols);

	LPDIRECT3DVERTEXBUFFER9 GetD3DVB();
	LPDIRECT3DINDEXBUFFER9 GetD3DIB();

	bool UsingFastLookup( void )		{ return m_bUseFastNormalLookup; }	//Using a lookup table to average adjacent triangle face normals?
	void SetUseFastLookup( void )		{ m_bUseFastNormalLookup = true; }
	void ResetUseFastLookup( void )		{ m_bUseFastNormalLookup = false; }
	void CopyVBToRenderVB( void );
	int  GetCurrentNumRows( void )		{ return m_iNumRows; }
	int  GetCurrentNumCols( void )		{ return m_iNumCols; }
	void SetFirstTime(void)				{ m_bFirstDerivativeThru = true; m_bFirstNormalLookupThru = true; }
	float GetMaxElevation( void );

	FM_GET_SET_ACCESSOR( double, H );
	FM_GET_SET_ACCESSOR( double, Lacunarity );
	FM_GET_SET_ACCESSOR( double, Octaves );
	FM_GET_SET_ACCESSOR( double, Offset );
	FM_GET_SET_ACCESSOR( double, Gain );

private:
	void GenVB();
	void GenIB();
	void GenNormalLookupTable();
	void GenVertexNeighborList();
	void InitNeighborList();
	void InitVertexNeighborList();
	void CreateNeighborList();
	void AddToNeighborList(int me, int myNeighbor);
	int IsNeighborOf(int a, int b);
	void LoadD3D_VB();
	void LoadD3D_IB();
	void PrintD3D_VB();
	void PrintD3D_IB();
	void ReloadVB();
	void ReloadIB();
	void DeleteMesh();

	double fBm( float vertex[], double H, double lacunarity, double octaves );
	double HybridMultiFractal( float vertex[], double H, double lacunarity, double octaves, double offset );
	double RidgedMultiFractal( float vertex[], double H, double lacunarity, double octaves, double offset, double gain );

	bool m_bFirstDerivativeThru;
	bool m_bFirstNormalLookupThru;

	// Number of vertices in a row == iNumCols
	int m_iNumCols;
	// Number of vertices in a col == iNumRows
	int m_iNumRows;

	FRACTAL_ALGORITHM m_fractalAlgorithm;
	double * m_FractalExponentArray;
	bool m_firstFractalCalc;

	short int *m_pIB;

	// Lookup table for the normals which will be averaged for each vertex
	int *m_pNormalLookupTable;

	CUSTOM_VERT_POS_NORMAL_TEXTURE0 *m_pVB;
	int *m_pNeighborList;
	CVector3 *m_pFaceNormals;

	// Controls how vertex normals are calculated
	bool m_bUseFastNormalLookup;

	LPDIRECT3DVERTEXBUFFER9 m_pVBD3D;
	LPDIRECT3DINDEXBUFFER9 m_pIBD3D;
	IDirect3DDevice9 *m_pDevice;

	float m_fMaxElevation;

	double m_H;							// Hurst index - in mathematical literature classifies the fBm and dictates fractal dimension
	double m_Lacunarity;				// Lacunarity dictates the gap between successive frequencies
	double m_Octaves;					// Octaves dicates the number of frequencies in the fBm
	double m_Offset;					// Offset dictates the offset from the lowest elevation and determines "multifractality"
	double m_Gain;						// Controls the amplitude of the frequency
};

#endif