//-------------------------------------------------------------------------------------
//
// Copyright 2007 Intel Corporation
// All Rights Reserved
//
// Permission is granted to use, copy, distribute and prepare derivative works of this
// software for any purpose and without fee, provided, that the above copyright notice
// and this statement appear in all copies.  Intel makes no representations about the
// suitability of this software for any purpose.  THIS SOFTWARE IS PROVIDED "AS IS."
// INTEL SPECIFICALLY DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, AND ALL LIABILITY,
// INCLUDING CONSEQUENTIAL AND OTHER INDIRECT DAMAGES, FOR THE USE OF THIS SOFTWARE,
// INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PROPRIETARY RIGHTS, AND INCLUDING THE
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  Intel does not
// assume any responsibility for any errors which may appear in this software nor any
// responsibility to update it.
//
//-------------------------------------------------------------------------------------

//--------------------------------------------------------------------------------------
// File: FractalTerrain.cpp
//
// This sample shows a simple example of the Microsoft Direct3D's High-Level
// Shader Language (HLSL) using the Effect interface.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//--------------------------------------------------------------------------------------
#include "DXUT.h"
#include "DXUTcamera.h"
#include "DXUTgui.h"
#include "DXUTsettingsdlg.h"
#include "SDKmisc.h"
#include "resource.h"
#include <assert.h>
#include "FractalMesh.h"

using namespace std;

//#define DEBUG_VS   // Uncomment this line to debug vertex shaders
//#define DEBUG_PS   // Uncomment this line to debug pixel shaders


//--------------------------------------------------------------------------------------
// Global variables
//--------------------------------------------------------------------------------------
ID3DXFont*              g_pFont = NULL;         // Font for drawing text
ID3DXSprite*            g_pSprite = NULL;       // Sprite for batching draw text calls
bool                    g_bShowHelp = true;     // If true, it renders the UI control text
CModelViewerCamera      g_Camera;               // A model viewing camera
ID3DXEffect*            g_pEffect = NULL;       // D3DX effect interface
ID3DXMesh*              g_pMesh = NULL;         // Mesh object

IDirect3DTexture9*      g_pSandTexture = NULL;  // Mesh texture
IDirect3DTexture9*      g_pGrassTexture = NULL;  // Mesh texture
IDirect3DTexture9*      g_pRockTexture = NULL;  // Mesh texture
IDirect3DTexture9*      g_pSnowTexture = NULL;  // Mesh texture


CDXUTDialogResourceManager g_DialogResourceManager; // manager for shared resources of dialogs
CD3DSettingsDlg         g_SettingsDlg;          // Device settings dialog
CDXUTDialog             g_HUD;                  // manages the 3D UI
CDXUTDialog             g_SampleUI;             // dialog for sample specific controls
CDXUTDialog				g_WaveSpecificUI;		// dialog for wave specific controls
bool					g_bEnablePreshader;
bool                    g_bRenderWireframe;
D3DXMATRIXA16           g_mCenterWorld;

// Lighting
#define MAX_LIGHTS 3
CDXUTDirectionWidget	g_LightControl[MAX_LIGHTS];
float					g_fLightScale;
int						g_nNumActiveLights;
int						g_nActiveLight;

// FractalMesh
FRACTAL_ALGORITHM		g_eFractalAlgorithm = FBM;
FractalMesh*			g_pTerrainMesh = NULL;
int						g_iNumRows = 90;
int						g_iNumCols = 90;
LPDIRECT3DVERTEXBUFFER9	g_pTerrainMeshVB = NULL;
LPDIRECT3DINDEXBUFFER9	g_pTerrainMeshIB = NULL;
long					lThreadLaunched;
bool					bUseWaveDerivative;
bool					bUseFastNormalLookup;
double					g_H;
double					g_lacunarity;
double					g_octaves;
double					g_double;
double					g_offset;
double					g_gain;

//--------------------------------------------------------------------------------------
// UI control IDs
//--------------------------------------------------------------------------------------
#define IDC_TOGGLEFULLSCREEN			1
#define IDC_CHANGEDEVICE				4
#define IDC_RENDER_WIREFRAME			5
#define IDC_NUM_LIGHTS					6
#define IDC_NUM_LIGHTS_STATIC			7
#define IDC_ACTIVE_LIGHT				8
#define IDC_LIGHT_SCALE					9
#define IDC_LIGHT_SCALE_STATIC			10
#define IDC_FRACTAL_ALGORITHM			11
#define IDC_H_SCALE_STATIC				12
#define IDC_H_SCALE						13
#define IDC_LACUNARITY_SCALE_STATIC		14
#define IDC_LACUNARITY_SCALE			15
#define IDC_OCTAVES_SCALE_STATIC		16
#define IDC_OCTAVES_SCALE				17
#define IDC_OFFSET_SCALE_STATIC			18
#define IDC_OFFSET_SCALE				19
#define IDC_GAIN_SCALE_STATIC			20
#define IDC_GAIN_SCALE					21

//--------------------------------------------------------------------------------------
// UI control data
//--------------------------------------------------------------------------------------
#define NUM_FRACTAL_ALGORITHM_NAMES 3
const TCHAR *FRACTAL_ALGORITHM_NAMES[] = 
{
	L"fBm",
	L"Hybrid Multifractal",
	L"Ridged Multifractal"
};

#define DOUBLE_SCALAR	1000.0f
#define MAX_H			1000
#define MAX_LACUNARITY	5000
#define MAX_OCTAVES		7000
#define MAX_OFFSET		1000
#define MAX_GAIN		1000

//--------------------------------------------------------------------------------------
// Forward declarations
//--------------------------------------------------------------------------------------
bool    CALLBACK IsDeviceAcceptable( D3DCAPS9* pCaps, D3DFORMAT AdapterFormat, D3DFORMAT BackBufferFormat, bool bWindowed, void* pUserContext );
bool    CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, void* pUserContext );
HRESULT CALLBACK OnCreateDevice( IDirect3DDevice9* pd3dDevice, const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext );
HRESULT CALLBACK OnResetDevice( IDirect3DDevice9* pd3dDevice, const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext );
void    CALLBACK OnFrameMove( double fTime, float fElapsedTime, void* pUserContext );
void    CALLBACK OnFrameRender( IDirect3DDevice9* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext );
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing, void* pUserContext );
void    CALLBACK KeyboardProc( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext );
void    CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext );
void    CALLBACK OnLostDevice( void* pUserContext );
void    CALLBACK OnDestroyDevice( void* pUserContext );

void    InitApp();
HRESULT LoadMesh( IDirect3DDevice9* pd3dDevice, WCHAR* strFileName, ID3DXMesh** ppMesh );
void    RenderText( double fTime );
void	UpdateFractalControls( FRACTAL_ALGORITHM selectedAlgorithm );
void	SetDefaultFractalSettings();
void	UpdateFractalMesh( FRACTAL_ALGORITHM newAlgorithm );

//--------------------------------------------------------------------------------------
// Entry point to the program. Initializes everything and goes into a message processing
// loop. Idle time is used to render the scene.
//--------------------------------------------------------------------------------------
INT WINAPI wWinMain( HINSTANCE, HINSTANCE, LPWSTR, int )
{
	// Enable run-time memory check for debug builds.
#if defined(DEBUG) | defined(_DEBUG)
	_CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif

	// Set the callback functions. These functions allow DXUT to notify
	// the application about device changes, user input, and windows messages.  The
	// callbacks are optional so you need only set callbacks for events you're interested
	// in. However, if you don't handle the device reset/lost callbacks then the sample
	// framework won't be able to reset your device since the application must first
	// release all device resources before resetting.  Likewise, if you don't handle the
	// device created/destroyed callbacks then DXUT won't be able to
	// recreate your device resources.
	DXUTSetCallbackD3D9DeviceAcceptable( IsDeviceAcceptable );
	DXUTSetCallbackD3D9DeviceCreated( OnCreateDevice );
	DXUTSetCallbackD3D9DeviceReset( OnResetDevice );
	DXUTSetCallbackD3D9FrameRender( OnFrameRender );
	DXUTSetCallbackD3D9DeviceLost( OnLostDevice );
	DXUTSetCallbackD3D9DeviceDestroyed( OnDestroyDevice );
	DXUTSetCallbackMsgProc( MsgProc );
	DXUTSetCallbackKeyboard( KeyboardProc );
	DXUTSetCallbackFrameMove( OnFrameMove );
	DXUTSetCallbackDeviceChanging( ModifyDeviceSettings );

	// Show the cursor and clip it when in full screen
	DXUTSetCursorSettings( true, true );

	InitApp();

	// Initialize DXUT and create the desired Win32 window and Direct3D
	// device for the application. Calling each of these functions is optional, but they
	// allow you to set several options which control the behavior of the framework.
	DXUTInit( true, true ); // Parse the command line and show msgboxes
	DXUTSetHotkeyHandling( true, true, true );
	DXUTCreateWindow( L"FractalTerrain" );
	DXUTCreateDevice( true, 640, 480 );

	// Pass control to DXUT for handling the message pump and
	// dispatching render calls. DXUT will call your FrameMove
	// and FrameRender callback when there is idle time between handling window messages.
	DXUTMainLoop();

	// Perform any application-level cleanup here. Direct3D device resources are released within the
	// appropriate callback functions and therefore don't require any cleanup code here.

	return DXUTGetExitCode();
}

//--------------------------------------------------------------------------------------
// Sets defaults for the fractal algorithms based on aestehtically pleasing and interesting results
//	per the goals of that algorithm.  A notable exception here is the Ridged Multifractal settings
//	which created an itneresting non-ridged result at this offset creating volcanic peaks interspersed
//	with rolling hills.
//--------------------------------------------------------------------------------------
void SetDefaultFractalSettings()
{
	switch (g_eFractalAlgorithm)
	{
	case FBM:
		{
			g_H = 0.7601f;
			g_lacunarity = 0.6019f;
			g_octaves = 5.0f;
			g_offset = 0.0f;
			g_gain = 0.0f;
			break;
		}
	case HYBRIDMULTIFRACTAL:
		{
			g_H = 0.5f;
			g_lacunarity = 0.551f;
			g_octaves = 3.571f;
			g_offset = 0.001f;
			g_gain = 0.0f;
			break;
		}
	case RIDGEDMULTIFRACTAL:
		{
			g_H = 0.52587983568f;
			g_lacunarity = 1.001f;
			g_octaves = 3.850f;
			g_offset = 0.011f;
			g_gain = 0.491f;
			break;
		}
	default:
		{
			assert( false );
			break;
		}
	}
}

//--------------------------------------------------------------------------------------
// Update UI controls based on the selected Fractal algorithm
//--------------------------------------------------------------------------------------
void UpdateFractalControls()
{
	CDXUTSlider *pHSlider = g_SampleUI.GetSlider( IDC_H_SCALE );
	CDXUTSlider *pLacunaritySlider = g_SampleUI.GetSlider( IDC_LACUNARITY_SCALE );
	CDXUTSlider *pOctavesSlider = g_SampleUI.GetSlider( IDC_OCTAVES_SCALE );
	CDXUTSlider *pOffsetSlider = g_SampleUI.GetSlider( IDC_OFFSET_SCALE );
	CDXUTSlider *pGainSlider = g_SampleUI.GetSlider( IDC_GAIN_SCALE );

	switch (g_eFractalAlgorithm)
	{
	case FBM:
		{
			pHSlider->SetEnabled( true );
			pHSlider->SetValue( (int)(g_H*DOUBLE_SCALAR) );
			pLacunaritySlider->SetEnabled( true );
			pLacunaritySlider->SetValue( (int)(g_lacunarity*DOUBLE_SCALAR) );
			pOctavesSlider->SetEnabled( true );
			pOctavesSlider->SetValue( (int)(g_octaves*DOUBLE_SCALAR) );
			pOffsetSlider->SetEnabled( false );
			pOffsetSlider->SetValue( (int)(g_offset*DOUBLE_SCALAR) );
			pGainSlider->SetEnabled( false );
			pGainSlider->SetValue( (int)(g_gain*DOUBLE_SCALAR) );
			break;
		}
	case HYBRIDMULTIFRACTAL:
		{
			pHSlider->SetEnabled( true );
			pHSlider->SetValue( (int)(g_H*DOUBLE_SCALAR) );
			pLacunaritySlider->SetEnabled( true );
			pLacunaritySlider->SetValue( (int)(g_lacunarity*DOUBLE_SCALAR) );
			pOctavesSlider->SetEnabled( true );
			pOctavesSlider->SetValue( (int)(g_octaves*DOUBLE_SCALAR) );
			pOffsetSlider->SetEnabled( true );
			pOffsetSlider->SetValue( (int)(g_offset*DOUBLE_SCALAR) );
			pGainSlider->SetEnabled( false );
			pGainSlider->SetValue( (int)(g_gain*DOUBLE_SCALAR) );
			break;
		}
	case RIDGEDMULTIFRACTAL:
		{
			pHSlider->SetEnabled( true );
			pHSlider->SetValue( (int)(g_H*DOUBLE_SCALAR) );
			pLacunaritySlider->SetEnabled( true );
			pLacunaritySlider->SetValue( (int)(g_lacunarity*DOUBLE_SCALAR) );
			pOctavesSlider->SetEnabled( true );
			pOctavesSlider->SetValue( (int)(g_octaves*DOUBLE_SCALAR) );
			pOffsetSlider->SetEnabled( true );
			pOffsetSlider->SetValue( (int)(g_offset*DOUBLE_SCALAR) );
			pGainSlider->SetEnabled( true );
			pGainSlider->SetValue( (int)(g_gain*DOUBLE_SCALAR) );
			break;
		}
	default:
		{
			assert( false );
			break;
		}
	}

	WCHAR sz[100];
	g_H = ((double)g_SampleUI.GetSlider( IDC_H_SCALE )->GetValue())/DOUBLE_SCALAR;
	StringCchPrintf( sz, 100, L"H: %0.3f", g_H );
	g_SampleUI.GetStatic( IDC_H_SCALE_STATIC )->SetText( sz );

	g_lacunarity = ((double)g_SampleUI.GetSlider( IDC_LACUNARITY_SCALE )->GetValue())/DOUBLE_SCALAR;
	StringCchPrintf( sz, 100, L"Lacunarity: %0.3f", g_lacunarity );
	g_SampleUI.GetStatic( IDC_LACUNARITY_SCALE_STATIC )->SetText( sz );

	g_octaves = ((double)g_SampleUI.GetSlider( IDC_OCTAVES_SCALE )->GetValue())/DOUBLE_SCALAR;
	StringCchPrintf( sz, 100, L"Octaves: %0.3f", g_octaves );
	g_SampleUI.GetStatic( IDC_OCTAVES_SCALE_STATIC )->SetText( sz );

	g_offset = ((double)g_SampleUI.GetSlider( IDC_OFFSET_SCALE )->GetValue())/DOUBLE_SCALAR;
	StringCchPrintf( sz, 100, L"Offset: %0.3f", g_offset );
	g_SampleUI.GetStatic( IDC_OFFSET_SCALE_STATIC )->SetText( sz );

	g_gain = ((double)g_SampleUI.GetSlider( IDC_GAIN_SCALE )->GetValue())/DOUBLE_SCALAR;
	StringCchPrintf( sz, 100, L"Gain: %0.3f", g_gain );
	g_SampleUI.GetStatic( IDC_GAIN_SCALE_STATIC )->SetText( sz );
}

//--------------------------------------------------------------------------------------
// Recompute the terrain mesh with values from the GUI controls
//--------------------------------------------------------------------------------------
void UpdateFractalMesh( FRACTAL_ALGORITHM fractalAlgorithm )
{
	assert( g_pTerrainMesh );

	g_pTerrainMesh->SetH( g_H);
	g_pTerrainMesh->SetLacunarity( g_lacunarity );
	g_pTerrainMesh->SetOctaves( g_octaves );
	g_pTerrainMesh->SetOffset( g_offset );
	g_pTerrainMesh->SetGain( g_gain );
	g_pTerrainMesh->ReComputeTerrain(fractalAlgorithm);
}


//--------------------------------------------------------------------------------------
// Initialize the app
//--------------------------------------------------------------------------------------
void InitApp()
{
	g_bEnablePreshader = true;
	g_bRenderWireframe = false;

	// Readjustment for lighting for ideal display characteristics at the default viewing perspective
	for( int i=0; i<MAX_LIGHTS; i++ )
		g_LightControl[i].SetLightDirection( D3DXVECTOR3( sinf(D3DX_PI*2*i/MAX_LIGHTS-D3DX_PI/6), -sinf(D3DX_PI*2*i/MAX_LIGHTS-D3DX_PI/6)+.2f, -cosf(D3DX_PI*2*i/MAX_LIGHTS-D3DX_PI/6) ) );

	g_nActiveLight = 0;
	g_nNumActiveLights = 1;
	g_fLightScale = 1.0f;

	// Initialize dialogs
	g_SettingsDlg.Init( &g_DialogResourceManager );
	g_HUD.Init( &g_DialogResourceManager );
	g_SampleUI.Init( &g_DialogResourceManager );

	int iY = 10;

	g_HUD.SetCallback( OnGUIEvent );
	g_HUD.AddButton( IDC_TOGGLEFULLSCREEN, L"Toggle full screen", 35, iY, 125, 22 );
	g_HUD.AddButton( IDC_CHANGEDEVICE, L"Change device (F2)", 35, iY += 24, 125, 22, VK_F2 );

	WCHAR sz[100];

	g_SampleUI.SetCallback( OnGUIEvent );

	iY = -110;

	CDXUTComboBox *pFractalComboBox = NULL;
	g_SampleUI.AddComboBox( IDC_FRACTAL_ALGORITHM, 0, iY += 10, 160, 24, 'R', false, &pFractalComboBox );
	pFractalComboBox->SetDropHeight( 30 );

	for( int items = 0; items < NUM_FRACTAL_ALGORITHM_NAMES; items++ )
	{
		pFractalComboBox->AddItem( FRACTAL_ALGORITHM_NAMES[items], IntToPtr(items) );
	}

	iY += 10;
	StringCchPrintf( sz, 100, L"H: %0.3f", g_H );
	g_SampleUI.AddStatic( IDC_H_SCALE_STATIC, sz, 35, iY += 18, 125, 22 );
	g_SampleUI.AddSlider( IDC_H_SCALE, 50, iY += 18, 100, 22, 1, MAX_H, (int)(g_H*DOUBLE_SCALAR));

	iY += 5;
	StringCchPrintf( sz, 100, L"Lacunarity: %0.3f", g_lacunarity );
	g_SampleUI.AddStatic( IDC_LACUNARITY_SCALE_STATIC, sz, 35, iY += 18, 125, 22 );
	g_SampleUI.AddSlider( IDC_LACUNARITY_SCALE, 50, iY += 18, 100, 22, 1, MAX_LACUNARITY, (int)(g_lacunarity*DOUBLE_SCALAR));

	iY += 5;
	StringCchPrintf( sz, 100, L"Octaves: %0.3f", g_octaves );
	g_SampleUI.AddStatic( IDC_OCTAVES_SCALE_STATIC, sz, 35, iY += 18, 125, 22 );
	g_SampleUI.AddSlider( IDC_OCTAVES_SCALE, 50, iY += 18, 100, 22, 1, MAX_OCTAVES, (int)(g_octaves*DOUBLE_SCALAR));

	iY += 5;
	StringCchPrintf( sz, 100, L"Offset: %0.3f", g_offset );
	g_SampleUI.AddStatic( IDC_OFFSET_SCALE_STATIC, sz, 35, iY += 18, 125, 22 );
	g_SampleUI.AddSlider( IDC_OFFSET_SCALE, 50, iY += 18, 100, 22, 1, MAX_OFFSET, (int)(g_offset*DOUBLE_SCALAR));

	iY += 5;
	StringCchPrintf( sz, 100, L"Gain: %0.3f", g_gain );
	g_SampleUI.AddStatic( IDC_GAIN_SCALE_STATIC, sz, 35, iY += 18, 125, 22 );
	g_SampleUI.AddSlider( IDC_GAIN_SCALE, 50, iY += 18, 100, 22, 1, MAX_GAIN, (int)(g_gain*DOUBLE_SCALAR));

	iY += 20;
	StringCchPrintf( sz, 100, L"# Lights: %d", g_nNumActiveLights );
	g_SampleUI.AddStatic( IDC_NUM_LIGHTS_STATIC, sz, 35, iY += 18, 125, 22 );
	g_SampleUI.AddSlider( IDC_NUM_LIGHTS, 50, iY += 18, 100, 22, 1, MAX_LIGHTS, g_nNumActiveLights );

	iY += 5;
	StringCchPrintf( sz, 100, L"Light scale: %0.2f", g_fLightScale );
	g_SampleUI.AddStatic( IDC_LIGHT_SCALE_STATIC, sz, 35, iY += 18, 125, 22 );
	g_SampleUI.AddSlider( IDC_LIGHT_SCALE, 50, iY += 18, 100, 22, 0, 20, (int) (g_fLightScale * 10.0f) );

	iY += 5;
	g_SampleUI.AddButton( IDC_ACTIVE_LIGHT, L"Change active light (K)", 35, iY += 20, 125, 22, 'K' );
	g_SampleUI.AddCheckBox( IDC_RENDER_WIREFRAME, L"Render wireframe", 35, iY += 24, 125, 22, g_bRenderWireframe );

	g_eFractalAlgorithm = FBM;
	SetDefaultFractalSettings();
	UpdateFractalControls();
}


//--------------------------------------------------------------------------------------
// Called during device initialization, this code checks the device for some
// minimum set of capabilities, and rejects those that don't pass by returning E_FAIL.
//--------------------------------------------------------------------------------------
bool CALLBACK IsDeviceAcceptable( D3DCAPS9* pCaps, D3DFORMAT AdapterFormat,
								 D3DFORMAT BackBufferFormat, bool bWindowed, void* pUserContext )
{
	// No fallback defined by this app, so reject any device that
	// doesn't support at least ps3.0
	if( pCaps->PixelShaderVersion < D3DPS_VERSION(3,0) )
		return false;

	// Skip backbuffer formats that don't support alpha blending
	IDirect3D9* pD3D = DXUTGetD3D9Object();
	if( FAILED( pD3D->CheckDeviceFormat( pCaps->AdapterOrdinal, pCaps->DeviceType,
		AdapterFormat, D3DUSAGE_QUERY_POSTPIXELSHADER_BLENDING,
		D3DRTYPE_TEXTURE, BackBufferFormat ) ) )
		return false;

	return true;
}


//--------------------------------------------------------------------------------------
// This callback function is called immediately before a device is created to allow the
// application to modify the device settings. The supplied pDeviceSettings parameter
// contains the settings that the framework has selected for the new device, and the
// application can make any desired changes directly to this structure.  Note however that
// DXUT will not correct invalid device settings so care must be taken
// to return valid device settings, otherwise IDirect3D9::CreateDevice() will fail.
//--------------------------------------------------------------------------------------
bool CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, void* pUserContext )
{
	assert( DXUT_D3D9_DEVICE == pDeviceSettings->ver );

	HRESULT hr;
	IDirect3D9* pD3D = DXUTGetD3D9Object();
	D3DCAPS9 caps;

	V( pD3D->GetDeviceCaps( pDeviceSettings->d3d9.AdapterOrdinal,
		pDeviceSettings->d3d9.DeviceType,
		&caps ) );

	// If device doesn't support HW T&L or doesn't support 3.0 vertex shaders in HW
	// then switch to SWVP.
	if( ( caps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT) == 0 ||
		caps.VertexShaderVersion < D3DVS_VERSION(3,0) )
	{
		pDeviceSettings->d3d9.BehaviorFlags = D3DCREATE_SOFTWARE_VERTEXPROCESSING;

		::MessageBox( NULL, L"The video adapter must support hardware T&L and Shader Model 3.0 to run properly.\n\nThe software renderer will be used instead.  You may experience slow refresh rates or rendering speed as a result.", L"Fractal Terrain", MB_OK );
	}

	// Debugging vertex shaders requires either REF or software vertex processing
	// and debugging pixel shaders requires REF.
#ifdef DEBUG_VS
	if( pDeviceSettings->d3d9.DeviceType != D3DDEVTYPE_REF )
	{
		pDeviceSettings->d3d9.BehaviorFlags &= ~D3DCREATE_HARDWARE_VERTEXPROCESSING;
		pDeviceSettings->d3d9.BehaviorFlags &= ~D3DCREATE_PUREDEVICE;
		pDeviceSettings->d3d9.BehaviorFlags |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;
	}
#endif
#ifdef DEBUG_PS
	pDeviceSettings->d3d9.DeviceType = D3DDEVTYPE_REF;
#endif
	// For the first device created if its a REF device, optionally display a warning dialog box
	static bool s_bFirstTime = true;
	if( s_bFirstTime )
	{
		s_bFirstTime = false;
		if( pDeviceSettings->d3d9.DeviceType == D3DDEVTYPE_REF )
			DXUTDisplaySwitchingToREFWarning( pDeviceSettings->ver );
	}

	return true;
}


//--------------------------------------------------------------------------------------
// This callback function will be called immediately after the Direct3D device has been
// created, which will happen during application initialization and windowed/full screen
// toggles. This is the best location to create D3DPOOL_MANAGED resources since these
// resources need to be reloaded whenever the device is destroyed. Resources created
// here should be released in the OnDestroyDevice callback.
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnCreateDevice( IDirect3DDevice9* pd3dDevice, const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext )
{
	HRESULT hr;

	V_RETURN( g_DialogResourceManager.OnD3D9CreateDevice( pd3dDevice ) );
	V_RETURN( g_SettingsDlg.OnD3D9CreateDevice( pd3dDevice ) );
	// Initialize the font
	V_RETURN( D3DXCreateFont( pd3dDevice, 15, 0, FW_BOLD, 1, FALSE, DEFAULT_CHARSET,
		OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE,
		L"Arial", &g_pFont ) );

	// Load the mesh
	V_RETURN( LoadMesh( pd3dDevice, L"Cube\\Cube.x", &g_pMesh ) );

	D3DXVECTOR3* pData;
	D3DXVECTOR3 vCenter;
	FLOAT fObjectRadius;
	V( g_pMesh->LockVertexBuffer( 0, (LPVOID*) &pData ) );
	V( D3DXComputeBoundingSphere( pData, g_pMesh->GetNumVertices(), D3DXGetFVFVertexSize( g_pMesh->GetFVF() ), &vCenter, &fObjectRadius ) );
	V( g_pMesh->UnlockVertexBuffer() );


	assert( g_pTerrainMesh == NULL );
	g_pTerrainMesh = new FractalMesh( pd3dDevice, g_iNumRows, g_iNumCols, g_eFractalAlgorithm );
	g_pTerrainMeshVB = g_pTerrainMesh->GetD3DVB();
	g_pTerrainMeshIB = g_pTerrainMesh->GetD3DIB();

	UpdateFractalMesh( g_eFractalAlgorithm );

	D3DXMatrixTranslation( &g_mCenterWorld, -vCenter.x, -vCenter.y, -vCenter.z );
	D3DXMATRIXA16 m;
	D3DXMatrixRotationY( &m, D3DX_PI);
	g_mCenterWorld *= m;
	D3DXMatrixRotationX( &m, D3DX_PI / 2.0f );
	g_mCenterWorld *= m;

	V_RETURN( CDXUTDirectionWidget::StaticOnD3D9CreateDevice( pd3dDevice ) );
	for( int i=0; i<MAX_LIGHTS; i++ )
		g_LightControl[i].SetRadius( fObjectRadius );

	// Define DEBUG_VS and/or DEBUG_PS to debug vertex and/or pixel shaders with the
	// shader debugger. Debugging vertex shaders requires either REF or software vertex
	// processing, and debugging pixel shaders requires REF.  The
	// D3DXSHADER_FORCE_*_SOFTWARE_NOOPT flag improves the debug experience in the
	// shader debugger.  It enables source level debugging, prevents instruction
	// reordering, prevents dead code elimination, and forces the compiler to compile
	// against the next higher available software target, which ensures that the
	// unoptimized shaders do not exceed the shader model limitations.  Setting these
	// flags will cause slower rendering since the shaders will be unoptimized and
	// forced into software.  See the DirectX documentation for more information about
	// using the shader debugger.
	DWORD dwShaderFlags = D3DXFX_NOT_CLONEABLE;

#if defined( DEBUG ) || defined( _DEBUG )
	// Set the D3DXSHADER_DEBUG flag to embed debug information in the shaders.
	// Setting this flag improves the shader debugging experience, but still allows
	// the shaders to be optimized and to run exactly the way they will run in
	// the release configuration of this program.
	dwShaderFlags |= D3DXSHADER_DEBUG;
#endif

#ifdef DEBUG_VS
	dwShaderFlags |= D3DXSHADER_FORCE_VS_SOFTWARE_NOOPT;
#endif
#ifdef DEBUG_PS
	dwShaderFlags |= D3DXSHADER_FORCE_PS_SOFTWARE_NOOPT;
#endif

	// Preshaders are parts of the shader that the effect system pulls out of the
	// shader and runs on the host CPU. They should be used if you are GPU limited.
	// The D3DXSHADER_NO_PRESHADER flag disables preshaders.
	if( !g_bEnablePreshader )
		dwShaderFlags |= D3DXSHADER_NO_PRESHADER;

	// Read the D3DX effect file
	WCHAR str[MAX_PATH];
	V_RETURN( DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"FractalTerrain.fx" ) );

	// If this fails, there should be debug output as to
	// why the .fx file failed to compile
	V_RETURN( D3DXCreateEffectFromFile( pd3dDevice, str, NULL, NULL, dwShaderFlags, NULL, &g_pEffect, NULL ) );

	// Sand
	V_RETURN( DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"Textures\\Sand.jpg" ) );

	V_RETURN( D3DXCreateTextureFromFileEx( pd3dDevice, str, D3DX_DEFAULT, D3DX_DEFAULT,
		D3DX_DEFAULT, 0, D3DFMT_UNKNOWN, D3DPOOL_MANAGED,
		D3DX_DEFAULT, D3DX_DEFAULT, 0,
		NULL, NULL, &g_pSandTexture ) );

	// Grass
	V_RETURN( DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"Textures\\Grass.jpg" ) );

	V_RETURN( D3DXCreateTextureFromFileEx( pd3dDevice, str, D3DX_DEFAULT, D3DX_DEFAULT,
		D3DX_DEFAULT, 0, D3DFMT_UNKNOWN, D3DPOOL_MANAGED,
		D3DX_DEFAULT, D3DX_DEFAULT, 0,
		NULL, NULL, &g_pGrassTexture ) );

	// Rock
	V_RETURN( DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"Textures\\Rock.jpg" ) );

	V_RETURN( D3DXCreateTextureFromFileEx( pd3dDevice, str, D3DX_DEFAULT, D3DX_DEFAULT,
		D3DX_DEFAULT, 0, D3DFMT_UNKNOWN, D3DPOOL_MANAGED,
		D3DX_DEFAULT, D3DX_DEFAULT, 0,
		NULL, NULL, &g_pRockTexture ) );

	// Snow
	V_RETURN( DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"Textures\\Snow.jpg" ) );

	V_RETURN( D3DXCreateTextureFromFileEx( pd3dDevice, str, D3DX_DEFAULT, D3DX_DEFAULT,
		D3DX_DEFAULT, 0, D3DFMT_UNKNOWN, D3DPOOL_MANAGED,
		D3DX_DEFAULT, D3DX_DEFAULT, 0,
		NULL, NULL, &g_pSnowTexture ) );

	// Set effect variables as needed
	D3DXCOLOR colorMtrlDiffuse(1.0f, 1.0f, 1.0f, 1.0f);
	D3DXCOLOR colorMtrlAmbient(0.35f, 0.35f, 0.35f, 0);

	V_RETURN( g_pEffect->SetValue( "g_MaterialAmbientColor", &colorMtrlAmbient, sizeof(D3DXCOLOR) ) );
	V_RETURN( g_pEffect->SetValue( "g_MaterialDiffuseColor", &colorMtrlDiffuse, sizeof(D3DXCOLOR) ) );

	V_RETURN( g_pEffect->SetTexture( "g_SandTexture", g_pSandTexture) );
	V_RETURN( g_pEffect->SetTexture( "g_GrassTexture", g_pGrassTexture) );
	V_RETURN( g_pEffect->SetTexture( "g_RockTexture", g_pRockTexture) );
	V_RETURN( g_pEffect->SetTexture( "g_SnowTexture", g_pSnowTexture) );

	// Setup the camera's view parameters
	D3DXVECTOR3 vecEye(0.0f, 5.0f, -15.0f); // Slightly tilt the eye location for a better view of the terrain
	D3DXVECTOR3 vecAt (0.0f, 0.0f, -0.0f);
	g_Camera.SetViewParams( &vecEye, &vecAt );
	g_Camera.SetRadius( fObjectRadius*3.0f, fObjectRadius*0.5f, fObjectRadius*10.0f );

	return S_OK;
}

//--------------------------------------------------------------------------------------
// This function loads the mesh and ensures the mesh has normals; it also optimizes the
// mesh for the graphics card's vertex cache, which improves performance by organizing
// the internal triangle list for less cache misses.
//--------------------------------------------------------------------------------------
HRESULT LoadMesh( IDirect3DDevice9* pd3dDevice, WCHAR* strFileName, ID3DXMesh** ppMesh )
{
	ID3DXMesh* pMesh = NULL;
	WCHAR str[MAX_PATH];
	HRESULT hr;

	// Load the mesh with D3DX and get back a ID3DXMesh*.  For this
	// sample we'll ignore the X file's embedded materials since we know
	// exactly the model we're loading.  See the mesh samples such as
	// "OptimizedMesh" for a more generic mesh loading example.
	V_RETURN( DXUTFindDXSDKMediaFileCch( str, MAX_PATH, strFileName ) );
	V_RETURN( D3DXLoadMeshFromX(str, D3DXMESH_MANAGED, pd3dDevice, NULL, NULL, NULL, NULL, &pMesh) );

	DWORD *rgdwAdjacency = NULL;

	// Make sure there are normals which are required for lighting
	if( !(pMesh->GetFVF() & D3DFVF_NORMAL) )
	{
		ID3DXMesh* pTempMesh;
		V( pMesh->CloneMeshFVF( pMesh->GetOptions(),
			pMesh->GetFVF() | D3DFVF_NORMAL,
			pd3dDevice, &pTempMesh ) );
		V( D3DXComputeNormals( pTempMesh, NULL ) );

		SAFE_RELEASE( pMesh );
		pMesh = pTempMesh;
	}

	// Optimize the mesh for this graphics card's vertex cache
	// so when rendering the mesh's triangle list the vertices will
	// cache hit more often so it won't have to re-execute the vertex shader
	// on those vertices so it will improve perf.
	rgdwAdjacency = new DWORD[pMesh->GetNumFaces() * 3];
	if( rgdwAdjacency == NULL )
		return E_OUTOFMEMORY;
	V( pMesh->GenerateAdjacency(1e-6f,rgdwAdjacency) );
	V( pMesh->OptimizeInplace(D3DXMESHOPT_VERTEXCACHE, rgdwAdjacency, NULL, NULL, NULL) );
	delete []rgdwAdjacency;

	*ppMesh = pMesh;

	return S_OK;
}


//--------------------------------------------------------------------------------------
// This callback function will be called immediately after the Direct3D device has been
// reset, which will happen after a lost device scenario. This is the best location to
// create D3DPOOL_DEFAULT resources since these resources need to be reloaded whenever
// the device is lost. Resources created here should be released in the OnLostDevice
// callback.
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnResetDevice( IDirect3DDevice9* pd3dDevice,
							   const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext )
{
	HRESULT hr;

	V_RETURN( g_DialogResourceManager.OnD3D9ResetDevice() );
	V_RETURN( g_SettingsDlg.OnD3D9ResetDevice() );

	if( g_pFont )
		V_RETURN( g_pFont->OnResetDevice() );
	if( g_pEffect )
		V_RETURN( g_pEffect->OnResetDevice() );

	// Create a sprite to help batch calls when drawing many lines of text
	V_RETURN( D3DXCreateSprite( pd3dDevice, &g_pSprite ) );

	if(g_pTerrainMesh == NULL)
	{
		g_pTerrainMesh = new FractalMesh(pd3dDevice, g_iNumRows, g_iNumCols, g_eFractalAlgorithm );
		g_pTerrainMeshVB = g_pTerrainMesh->GetD3DVB();
		g_pTerrainMeshIB = g_pTerrainMesh->GetD3DIB();

		UpdateFractalMesh( g_eFractalAlgorithm );
	}

	for( int i=0; i<MAX_LIGHTS; i++ )
		g_LightControl[i].OnD3D9ResetDevice( pBackBufferSurfaceDesc  );

	CDXUTComboBox *pFractalComboBox = g_SampleUI.GetComboBox( IDC_FRACTAL_ALGORITHM );
	pFractalComboBox->SetSelectedByData( (void *)g_eFractalAlgorithm );

	// Setup the camera's projection parameters
	float fAspectRatio = pBackBufferSurfaceDesc->Width / (FLOAT)pBackBufferSurfaceDesc->Height;
	g_Camera.SetProjParams( D3DX_PI/4, fAspectRatio, 2.0f, 4000.0f );
	g_Camera.SetWindow( pBackBufferSurfaceDesc->Width, pBackBufferSurfaceDesc->Height );
	g_Camera.SetButtonMasks( MOUSE_LEFT_BUTTON, MOUSE_WHEEL, MOUSE_MIDDLE_BUTTON );

	g_HUD.SetLocation( pBackBufferSurfaceDesc->Width-170, 0 );
	g_HUD.SetSize( 170, 170 );
	g_SampleUI.SetLocation( pBackBufferSurfaceDesc->Width-170, pBackBufferSurfaceDesc->Height-300 );
	g_SampleUI.SetSize( 170, 300 );

	return S_OK;
}


//--------------------------------------------------------------------------------------
// This callback function will be called once at the beginning of every frame. This is the
// best location for your application to handle updates to the scene, but is not
// intended to contain actual rendering calls, which should instead be placed in the
// OnFrameRender callback.
//--------------------------------------------------------------------------------------
void CALLBACK OnFrameMove( double fTime, float fElapsedTime, void* pUserContext )
{
	// Update the camera's position based on user input
	g_Camera.FrameMove( fElapsedTime );
}


//--------------------------------------------------------------------------------------
// This callback function will be called at the end of every frame to perform all the
// rendering calls for the scene, and it will also be called if the window needs to be
// repainted. After this function has returned, DXUT will call
// IDirect3DDevice9::Present to display the contents of the next buffer in the swap chain
//--------------------------------------------------------------------------------------
void CALLBACK OnFrameRender( IDirect3DDevice9* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext )
{
	// If the settings dialog is being shown, then
	// render it instead of rendering the app's scene
	if( g_SettingsDlg.IsActive() )
	{
		g_SettingsDlg.OnRender( fElapsedTime );
		return;
	}

	HRESULT hr;
	D3DXMATRIXA16 mWorldViewProjection;
	D3DXVECTOR3 vLightDir[MAX_LIGHTS];
	D3DXCOLOR   vLightDiffuse[MAX_LIGHTS];
	UINT iPass, cPasses;
	D3DXMATRIXA16 mWorld;
	D3DXMATRIXA16 mView;
	D3DXMATRIXA16 mProj;

	// Clear the render target and the zbuffer
	V( pd3dDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DXCOLOR(0.0f,0.25f,0.25f,0.55f), 1.0f, 0) );

	// Render the scene
	if( SUCCEEDED( pd3dDevice->BeginScene() ) )
	{
		// Get the projection & view matrix from the camera class
		mWorld = g_mCenterWorld * *g_Camera.GetWorldMatrix();
		mProj = *g_Camera.GetProjMatrix();
		mView = *g_Camera.GetViewMatrix();

		mWorldViewProjection = mWorld * mView * mProj;

		// Render the light arrow so the user can visually see the light dir
		for( int i=0; i<g_nNumActiveLights; i++ )
		{
			D3DXCOLOR arrowColor = ( i == g_nActiveLight ) ? D3DXCOLOR(1,1,0,1) : D3DXCOLOR(1,1,1,1);
			V( g_LightControl[i].OnRender9( arrowColor, &mView, &mProj, g_Camera.GetEyePt() ) );
			vLightDir[i] = g_LightControl[i].GetLightDirection();
			vLightDiffuse[i] = g_fLightScale * D3DXCOLOR(1,1,1,1);
		}

		V( g_pEffect->SetValue( "g_LightDir", vLightDir, sizeof(D3DXVECTOR3)*MAX_LIGHTS ) );
		V( g_pEffect->SetValue( "g_LightDiffuse", vLightDiffuse, sizeof(D3DXVECTOR4)*MAX_LIGHTS ) );

		// Update the effect's variables.  Instead of using strings, it would
		// be more efficient to cache a handle to the parameter by calling
		// ID3DXEffect::GetParameterByName
		V( g_pEffect->SetMatrix( "g_mWorldViewProjection", &mWorldViewProjection ) );
		V( g_pEffect->SetMatrix( "g_mWorld", &mWorld ) );
		V( g_pEffect->SetFloat( "g_fTime", (float)fTime ) );

		V( g_pEffect->SetFloat( "g_fTime", (float)fTime ) );
		V( g_pEffect->SetInt( "g_nNumLights", g_nNumActiveLights ) );


		V( g_pEffect->SetFloat( "g_fMaxElev", g_pTerrainMesh->GetMaxElevation() ) );

		// Render the scene with this technique
		// as defined in the .fx file
		D3DXCOLOR vDiffuseColor;
		if( g_bRenderWireframe )
		{
			vDiffuseColor = D3DXCOLOR(0.25f,0.45f,0.0f,1.0f);
			V( g_pEffect->SetTechnique( "RenderSceneNoTexture" ) );
		}
		else
		{
			vDiffuseColor = D3DXCOLOR(1.0f,1.0f,1.0f,1.0f);
			
			switch( g_nNumActiveLights )
			{
			case 1: V( g_pEffect->SetTechnique( "RenderSceneWithTexture1Light" ) ); break;
			case 2: V( g_pEffect->SetTechnique( "RenderSceneWithTexture2Light" ) ); break;
			case 3: V( g_pEffect->SetTechnique( "RenderSceneWithTexture3Light" ) ); break;
			}
		}

		V( g_pEffect->SetValue("g_MaterialDiffuseColor", &vDiffuseColor, sizeof(D3DXCOLOR) ) );

		// Apply the technique contained in the effect
		V( g_pEffect->Begin(&cPasses, 0) );

		for (iPass = 0; iPass < cPasses; iPass++)
		{
			V( g_pEffect->BeginPass(iPass) );

			// Render the fractal mesh
			g_pTerrainMesh->CopyVBToRenderVB();
			V( pd3dDevice->SetStreamSource(0, g_pTerrainMeshVB, 0, sizeof(CUSTOM_VERT_POS_NORMAL_TEXTURE0)) );
			pd3dDevice->SetFVF(D3D_FVF_CUSTOMVERTEX_POS_NORMAL_TEXTURE0);
			V( pd3dDevice->SetIndices(g_pTerrainMeshIB ) );
			V( pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, g_iNumRows*g_iNumCols, 0, (g_iNumRows-1)*(g_iNumCols-1)*2) );
			
			if( g_bRenderWireframe )
			{
				V( pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME ) );
			}
			else
			{
				V( pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID  ) );
			}
			// The effect interface queues up the changes and performs them
			// with the CommitChanges call. You do not need to call CommitChanges if
			// you are not setting any parameters between the BeginPass and EndPass.
			// V( g_pEffect->CommitChanges() );

			V( g_pEffect->EndPass() );
		}
		V( g_pEffect->End() );

		g_HUD.OnRender( fElapsedTime );
		g_SampleUI.OnRender( fElapsedTime );

		RenderText( fTime );

		V( pd3dDevice->EndScene() );
	}
}


//--------------------------------------------------------------------------------------
// Render the help and statistics text. This function uses the ID3DXFont interface for
// efficient text rendering.
//--------------------------------------------------------------------------------------
void RenderText( double fTime )
{
	// The helper object simply helps keep track of text position, and color
	// and then it calls pFont->DrawText( m_pSprite, strMsg, -1, &rc, DT_NOCLIP, m_clr );
	// If NULL is passed in as the sprite object, then it will work fine however the
	// pFont->DrawText() will not be batched together.  Batching calls will improves perf.
	CDXUTTextHelper txtHelper( g_pFont, g_pSprite, 15 );

	// Output statistics
	txtHelper.Begin();
	txtHelper.SetInsertionPos( 2, 0 );
	txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 0.0f, 1.0f ) );
	txtHelper.DrawTextLine( DXUTGetFrameStats( DXUTIsVsyncEnabled() ) );
	txtHelper.DrawTextLine( DXUTGetDeviceStats() );

	txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 1.0f, 1.0f ) );
	txtHelper.DrawFormattedTextLine( FRACTAL_ALGORITHM_NAMES[ g_eFractalAlgorithm] );

	// Draw help
	if( g_bShowHelp )
	{
		const D3DSURFACE_DESC* pd3dsdBackBuffer = DXUTGetD3D9BackBufferSurfaceDesc();
		txtHelper.SetInsertionPos( 2, pd3dsdBackBuffer->Height-15*6 );
		txtHelper.SetForegroundColor( D3DXCOLOR(1.0f, 0.75f, 0.0f, 1.0f ) );
		txtHelper.DrawTextLine( L"Controls:" );

		txtHelper.SetInsertionPos( 20, pd3dsdBackBuffer->Height-15*5 );
		txtHelper.DrawTextLine( L"Rotate model: Left mouse button\n"
			L"Rotate light: Right mouse button\n"
			L"Zoom camera: Mouse wheel scroll\n" );

		txtHelper.SetInsertionPos( 250, pd3dsdBackBuffer->Height-15*5 );
		txtHelper.DrawTextLine( L"Hide help: F1\n"
			L"Quit: ESC\n" );
	}
	else
	{
		txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 1.0f, 1.0f ) );
		txtHelper.DrawTextLine( L"Press F1 for help" );
	}
	txtHelper.End();
}


//--------------------------------------------------------------------------------------
// Before handling window messages, DXUT passes incoming windows
// messages to the application through this callback function. If the application sets
// *pbNoFurtherProcessing to TRUE, then DXUT will not process this message.
//--------------------------------------------------------------------------------------
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing, void* pUserContext )
{
	// Always allow dialog resource manager calls to handle global messages
	// so GUI state is updated correctly
	*pbNoFurtherProcessing = g_DialogResourceManager.MsgProc( hWnd, uMsg, wParam, lParam );
	if( *pbNoFurtherProcessing )
		return 0;

	if( g_SettingsDlg.IsActive() )
	{
		g_SettingsDlg.MsgProc( hWnd, uMsg, wParam, lParam );
		return 0;
	}

	// Give the dialogs a chance to handle the message first
	*pbNoFurtherProcessing = g_HUD.MsgProc( hWnd, uMsg, wParam, lParam );
	if( *pbNoFurtherProcessing )
		return 0;
	*pbNoFurtherProcessing = g_SampleUI.MsgProc( hWnd, uMsg, wParam, lParam );
	if( *pbNoFurtherProcessing )
		return 0;

	g_LightControl[g_nActiveLight].HandleMessages( hWnd, uMsg, wParam, lParam );

	// Pass all remaining windows messages to camera so it can respond to user input
	g_Camera.HandleMessages( hWnd, uMsg, wParam, lParam );

	return 0;
}


//--------------------------------------------------------------------------------------
// As a convenience, DXUT inspects the incoming windows messages for
// keystroke messages and decodes the message parameters to pass relevant keyboard
// messages to the application.  The framework does not remove the underlying keystroke
// messages, which are still passed to the application's MsgProc callback.
//--------------------------------------------------------------------------------------
void CALLBACK KeyboardProc( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext )
{
	if( bKeyDown )
	{
		switch( nChar )
		{
		case VK_F1: g_bShowHelp = !g_bShowHelp; break;
		}
	}
}


//--------------------------------------------------------------------------------------
// Handles the GUI events
//--------------------------------------------------------------------------------------
void CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext )
{
	WCHAR sz[100];

	switch( nControlID )
	{
	case IDC_TOGGLEFULLSCREEN: DXUTToggleFullScreen();
		break;

	case IDC_CHANGEDEVICE:     g_SettingsDlg.SetActive( !g_SettingsDlg.IsActive() );
		break;

	case IDC_FRACTAL_ALGORITHM:
		{
			CDXUTComboBox *pFractalComboBox = (CDXUTComboBox*)pControl;
			FRACTAL_ALGORITHM selectedFractalAlgorithm = (FRACTAL_ALGORITHM)(int)PtrToInt(pFractalComboBox->GetSelectedData());

			if( selectedFractalAlgorithm != g_eFractalAlgorithm )
			{
				// Order of operation counts here...
				g_eFractalAlgorithm = selectedFractalAlgorithm;
				SetDefaultFractalSettings();
				UpdateFractalMesh( selectedFractalAlgorithm );
				UpdateFractalControls();
			}

			break;
		}
	case IDC_H_SCALE:
		g_H = ((double)g_SampleUI.GetSlider( IDC_H_SCALE )->GetValue())/DOUBLE_SCALAR;
		StringCchPrintf( sz, 100, L"H: %0.3f", g_H );
		g_SampleUI.GetStatic( IDC_H_SCALE_STATIC )->SetText( sz );
		UpdateFractalMesh( g_eFractalAlgorithm );
		break;

	case IDC_LACUNARITY_SCALE:
		g_lacunarity = ((double)g_SampleUI.GetSlider( IDC_LACUNARITY_SCALE )->GetValue())/DOUBLE_SCALAR;
		StringCchPrintf( sz, 100, L"Lacunarity: %0.3f", g_lacunarity );
		g_SampleUI.GetStatic( IDC_LACUNARITY_SCALE_STATIC )->SetText( sz );
		UpdateFractalMesh( g_eFractalAlgorithm );
		break;

	case IDC_OCTAVES_SCALE:
		g_octaves = ((double)g_SampleUI.GetSlider( IDC_OCTAVES_SCALE )->GetValue())/DOUBLE_SCALAR;
		StringCchPrintf( sz, 100, L"Octaves: %0.3f", g_octaves );
		g_SampleUI.GetStatic( IDC_OCTAVES_SCALE_STATIC )->SetText( sz );
		UpdateFractalMesh( g_eFractalAlgorithm );
		break;

	case IDC_OFFSET_SCALE:
		g_offset = ((double)g_SampleUI.GetSlider( IDC_OFFSET_SCALE )->GetValue())/DOUBLE_SCALAR;
		StringCchPrintf( sz, 100, L"Offset: %0.3f", g_offset );
		g_SampleUI.GetStatic( IDC_OFFSET_SCALE_STATIC )->SetText( sz );
		UpdateFractalMesh( g_eFractalAlgorithm );
		break;

	case IDC_GAIN_SCALE:
		g_gain = ((double)g_SampleUI.GetSlider( IDC_GAIN_SCALE )->GetValue())/DOUBLE_SCALAR;
		StringCchPrintf( sz, 100, L"Gain: %0.3f", g_gain );
		g_SampleUI.GetStatic( IDC_GAIN_SCALE_STATIC )->SetText( sz );
		UpdateFractalMesh( g_eFractalAlgorithm );
		break;

	case IDC_RENDER_WIREFRAME:
		{
			g_bRenderWireframe = g_SampleUI.GetCheckBox( IDC_RENDER_WIREFRAME )->GetChecked();

			break;
		}

	case IDC_ACTIVE_LIGHT:
		if( !g_LightControl[g_nActiveLight].IsBeingDragged() )
		{
			g_nActiveLight++;
			g_nActiveLight %= g_nNumActiveLights;
		}
		break;

	case IDC_NUM_LIGHTS:
		if( !g_LightControl[g_nActiveLight].IsBeingDragged() )
		{
			WCHAR sz[100];
			StringCchPrintf( sz, 100, L"# Lights: %d", g_SampleUI.GetSlider( IDC_NUM_LIGHTS )->GetValue() );
			g_SampleUI.GetStatic( IDC_NUM_LIGHTS_STATIC )->SetText( sz );

			g_nNumActiveLights = g_SampleUI.GetSlider( IDC_NUM_LIGHTS )->GetValue();
			g_nActiveLight %= g_nNumActiveLights;
		}
		break;

	case IDC_LIGHT_SCALE:
		g_fLightScale = (float) (g_SampleUI.GetSlider( IDC_LIGHT_SCALE )->GetValue() * 0.10f);

		WCHAR sz[100];
		StringCchPrintf( sz, 100, L"Light scale: %0.2f", g_fLightScale );
		g_SampleUI.GetStatic( IDC_LIGHT_SCALE_STATIC )->SetText( sz );
		break;
	}
}


//--------------------------------------------------------------------------------------
// This callback function will be called immediately after the Direct3D device has
// entered a lost state and before IDirect3DDevice9::Reset is called. Resources created
// in the OnResetDevice callback should be released here, which generally includes all
// D3DPOOL_DEFAULT resources. See the "Lost Devices" section of the documentation for
// information about lost devices.
//--------------------------------------------------------------------------------------
void CALLBACK OnLostDevice( void* pUserContext )
{
	g_DialogResourceManager.OnD3D9LostDevice();
	g_SettingsDlg.OnD3D9LostDevice();
	CDXUTDirectionWidget::StaticOnD3D9LostDevice();
	if( g_pFont )
		g_pFont->OnLostDevice();
	if( g_pEffect )
		g_pEffect->OnLostDevice();

	if(g_pTerrainMesh)
	{
		g_pTerrainMesh->OnLostDevice();
		delete g_pTerrainMesh;
		g_pTerrainMesh = NULL;
		g_pTerrainMeshVB = NULL;
		g_pTerrainMeshIB = NULL;
	}

	SAFE_RELEASE(g_pSprite);
}


//--------------------------------------------------------------------------------------
// This callback function will be called immediately after the Direct3D device has
// been destroyed, which generally happens as a result of application termination or
// windowed/full screen toggles. Resources created in the OnCreateDevice callback
// should be released here, which generally includes all D3DPOOL_MANAGED resources.
//--------------------------------------------------------------------------------------
void CALLBACK OnDestroyDevice( void* pUserContext )
{
	g_DialogResourceManager.OnD3D9DestroyDevice();
	g_SettingsDlg.OnD3D9DestroyDevice();
	CDXUTDirectionWidget::StaticOnD3D9DestroyDevice();
	SAFE_RELEASE(g_pEffect);
	SAFE_RELEASE(g_pFont);
	SAFE_RELEASE(g_pMesh);

	SAFE_RELEASE(g_pSandTexture);
	SAFE_RELEASE(g_pGrassTexture);
	SAFE_RELEASE(g_pRockTexture);
	SAFE_RELEASE(g_pSnowTexture);
}