//-------------------------------------------------------------------------------------
//
// Copyright 2007 Intel Corporation
// All Rights Reserved
//
// Permission is granted to use, copy, distribute and prepare derivative works of this
// software for any purpose and without fee, provided, that the above copyright notice
// and this statement appear in all copies.  Intel makes no representations about the
// suitability of this software for any purpose.  THIS SOFTWARE IS PROVIDED "AS IS."
// INTEL SPECIFICALLY DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, AND ALL LIABILITY,
// INCLUDING CONSEQUENTIAL AND OTHER INDIRECT DAMAGES, FOR THE USE OF THIS SOFTWARE,
// INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PROPRIETARY RIGHTS, AND INCLUDING THE
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  Intel does not
// assume any responsibility for any errors which may appear in this software nor any
// responsibility to update it.
//
// Portions derived from the Microsoft DirectX SDK November 2007.
// Copyright (c) Microsoft Corporation. All rights reserved.
//-------------------------------------------------------------------------------------

//--------------------------------------------------------------------------------------
// File: FractalTerrain.fx
//
// The effect file for the FractalTerrain sample.  
// 
//--------------------------------------------------------------------------------------


//--------------------------------------------------------------------------------------
// Global variables
//--------------------------------------------------------------------------------------
float4 g_MaterialAmbientColor;      // Material's ambient color
float4 g_MaterialDiffuseColor;      // Material's diffuse color
int g_nNumLights;

float3 g_LightDir[3];               // Light's direction in world space
float4 g_LightDiffuse[3];           // Light's diffuse color
float4 g_LightAmbient;              // Light's ambient color

texture g_SandTexture;              // Color texture for mesh
texture g_GrassTexture;             // Color texture for mesh
texture g_RockTexture;              // Color texture for mesh
texture g_SnowTexture;              // Color texture for mesh

float    g_fTime;                   // App's time in seconds
float4x4 g_mWorld;                  // World matrix for object
float4x4 g_mWorldViewProjection;    // World * View * Projection matrix

float	g_fMaxElev;
float	g_fZonePlains = 0.0;
float	g_fZoneHills = 0.25;
float	g_fZoneAlpine = 0.70;
float	g_fZoneMountain = 1.0;


//--------------------------------------------------------------------------------------
// Texture samplers
//--------------------------------------------------------------------------------------
sampler SandTextureSampler = 
sampler_state
{
    Texture = <g_SandTexture>;
    MinFilter = Anisotropic;
    MagFilter = LINEAR;
    MipFilter = LINEAR;
    AddressU = WRAP;
    AddressV = WRAP;
};

sampler GrassTextureSampler = 
sampler_state
{
    Texture = <g_GrassTexture>;
    MinFilter = Anisotropic;
    MagFilter = LINEAR;
    MipFilter = LINEAR;
    AddressU = WRAP;
    AddressV = WRAP;
};

sampler RockTextureSampler = 
sampler_state
{
    Texture = <g_RockTexture>;
    MinFilter = Anisotropic;
    MagFilter = LINEAR;
    MipFilter = LINEAR;
    AddressU = WRAP;
    AddressV = WRAP;
};

sampler SnowTextureSampler = 
sampler_state
{
    Texture = <g_SnowTexture>;
    MinFilter = Anisotropic;
    MagFilter = LINEAR;
    MipFilter = LINEAR;
    AddressU = WRAP;
    AddressV = WRAP;
};

//--------------------------------------------------------------------------------------
// Vertex shader output structure
//--------------------------------------------------------------------------------------
struct VS_OUTPUT
{
    float4 Position   : POSITION;   // vertex position 
    float4 Diffuse    : COLOR0;     // vertex diffuse color (note that COLOR0 is clamped from 0..1)
    float2 TextureUV  : TEXCOORD0;  // vertex texture coords
    float  Elevation  : FOG;
};


//--------------------------------------------------------------------------------------
// This shader computes standard transform and lighting
//--------------------------------------------------------------------------------------
VS_OUTPUT RenderSceneVS( float4 vPos : POSITION, 
                         float3 vNormal : NORMAL,
                         uniform int nNumLights,
                         uniform bool bTexture )
{
    VS_OUTPUT Output;
    float3 vNormalWorldSpace;
  
    Output.Elevation = clamp(vPos.z / g_fMaxElev, 0.0f, 1.0f);
    
    // Transform the position from object space to homogeneous projection space
    Output.Position = mul(vPos, g_mWorldViewProjection);
    
    // Transform the normal from object space to world space    
    vNormalWorldSpace = normalize(mul(vNormal, (float3x3)g_mWorld)); // normal (world space)
    
    // Compute simple directional lighting equation
    float3 vTotalLightDiffuse = float3(0,0,0);
    for(int i=0; i<nNumLights; i++ )
        vTotalLightDiffuse += g_LightDiffuse[i] * max(0,dot(vNormalWorldSpace, g_LightDir[i]));
        
    Output.Diffuse.rgb = g_MaterialDiffuseColor * vTotalLightDiffuse + 
                         g_MaterialAmbientColor * g_LightAmbient;   
    Output.Diffuse.a = 1.0f; 
    
    if( bTexture )
    { 
        Output.TextureUV = float2(vPos.x * 1.0f, vPos.y * 1.0f);
    }
    else
    {
        Output.TextureUV = 0;
    }
    
    return Output;    
}


//--------------------------------------------------------------------------------------
// Pixel shader output structure
//--------------------------------------------------------------------------------------
struct PS_OUTPUT
{
    float4 RGBColor : COLOR0;  // Pixel color    
};

float4 TexturePresenceSand( in float2 texCoord,
                       in float lowB,
                       in float upB,
                       in float opt,
                       in float elev)
{
	if((elev < lowB) || (elev > upB))
	{
		return float4(0.0f, 0.0f, 0.0f, 0.0f);
	}

	float4 k = (elev <= opt) ? smoothstep(lowB, opt, elev) : smoothstep(upB, opt, elev);

	return k * tex2D( SandTextureSampler, texCoord );
}

float4 TexturePresenceGrass( in float2 texCoord,
                       in float lowB,
                       in float upB,
                       in float opt,
                       in float elev)
{
	if((elev < lowB) || (elev > upB))
	{
		return float4(0.0f, 0.0f, 0.0f, 0.0f);
	}

	float4 k = (elev <= opt) ? smoothstep(lowB, opt, elev) : smoothstep(upB, opt, elev);

	return k * tex2D( GrassTextureSampler, texCoord );
}

float4 TexturePresenceRock( in float2 texCoord,
                       in float lowB,
                       in float upB,
                       in float opt,
                       in float elev)
{
	if((elev < lowB) || (elev > upB))
	{
		return float4(0.0f, 0.0f, 0.0f, 0.0f);
	}

	float4 k = (elev <= opt) ? smoothstep(lowB, opt, elev) : smoothstep(upB, opt, elev);

	return k * tex2D(RockTextureSampler, texCoord );
}

float4 TexturePresenceSnow( in float2 texCoord,
                       in float lowB,
                       in float upB,
                       in float opt,
                       in float elev)
{
	if((elev < lowB) || (elev > upB))
	{
		return float4(0.0f, 0.0f, 0.0f, 0.0f);
	}

	float4 k = (elev <= opt) ? smoothstep(lowB, opt, elev) : smoothstep(upB, opt, elev);

	return k * tex2D(SnowTextureSampler, texCoord );
}

//--------------------------------------------------------------------------------------
// This shader outputs the pixel's color by modulating the texture's
//       color with diffuse material color
//--------------------------------------------------------------------------------------
PS_OUTPUT RenderScenePS( VS_OUTPUT In,
                         uniform bool bTexture ) 
{ 
    PS_OUTPUT Output;
	
    // Lookup mesh texture and modulate it with diffuse
    if( bTexture )
    {
		if (g_fMaxElev > 0.1f)
		{	
			Output.RGBColor = TexturePresenceSand( In.TextureUV,
	                                      -0.1f,
	                                      g_fZoneHills,
	                                      g_fZonePlains,
	                                      In.Elevation);
		
			Output.RGBColor += TexturePresenceGrass( In.TextureUV,
	                                      g_fZonePlains,
	                                      g_fZoneAlpine,
	                                      g_fZoneHills,
	                                      In.Elevation);
	                                                                           
			Output.RGBColor += TexturePresenceRock( In.TextureUV,
	                                      g_fZoneHills,
	                                      g_fZoneMountain,
	                                      g_fZoneAlpine,
	                                      In.Elevation);

			Output.RGBColor += TexturePresenceSnow( In.TextureUV,
	                                      g_fZoneAlpine,
	                                      100.1f,
	                                      g_fZoneMountain,
	                                      In.Elevation);
		}
		else
		{
			Output.RGBColor = TexturePresenceSand( In.TextureUV,
	                                      -0.1f,
	                                      g_fZoneHills,
	                                      g_fZonePlains,
	                                      In.Elevation);
		
			Output.RGBColor += TexturePresenceGrass( In.TextureUV,
	                                      g_fZonePlains,
	                                      100.1f,
	                                      g_fZoneHills,
	                                      In.Elevation);
		}
		
		Output.RGBColor *= In.Diffuse;
    }
    else
    {
        Output.RGBColor = In.Diffuse;
    }

    return Output;
}


//--------------------------------------------------------------------------------------
// Renders scene to render target
//--------------------------------------------------------------------------------------
technique RenderSceneWithTexture1Light
{
    pass P0
    {          
        VertexShader = compile vs_3_0 RenderSceneVS( 1, true );
        PixelShader  = compile ps_3_0 RenderScenePS( true );
    }
}

technique RenderSceneWithTexture2Light
{
    pass P0
    {          
        VertexShader = compile vs_3_0 RenderSceneVS( 2, true );
        PixelShader  = compile ps_3_0 RenderScenePS( true );
    }
}

technique RenderSceneWithTexture3Light
{
    pass P0
    {          
        VertexShader = compile vs_3_0 RenderSceneVS( 3, true );
        PixelShader  = compile ps_3_0 RenderScenePS( true );
    }
}

technique RenderSceneNoTexture
{
    pass P0
    {          
        VertexShader = compile vs_3_0 RenderSceneVS( 1, false );
        PixelShader  = compile ps_3_0 RenderScenePS( false );
    }
}
